/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_precon_create
 * lis_psolve
 * lis_psolvet
 ************************************************/
#undef __FUNC__
#define __FUNC__ "lis_precon_create_ilu"
int lis_precon_create_ilu(LIS_SOLVER solver, LIS_PRECON precon)
{
	int		k;
	int		err;

	LIS_DEBUG_FUNC_IN;
	k = solver->options[LIS_OPTIONS_FILL];
	if( k==0 )
	{
		err = lis_matrix_split(solver->A);
		if( err )
		{
			return err;
		}
		err = lis_precon_create_ilu0(solver,precon);
	}
	else
	{
		/*
		err = lis_precon_create_iluk(solver,precon);
		*/
		LIS_SETERR_IMP;
		return LIS_ERR_NOT_IMPLEMENTED;
	}
	LIS_DEBUG_FUNC_OUT;
	return err;
}

#undef __FUNC__
#define __FUNC__ "lis_precon_create_ilu0"
int lis_precon_create_ilu0(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				err;

	LIS_DEBUG_FUNC_IN;

	switch( solver->A->matrix_type )
	{
	case LIS_MATRIX_CRS:
		err = lis_precon_create_ilu0_crs(solver,precon);
		break;
	case LIS_MATRIX_BSR:
		err = lis_precon_create_ilu0_bsr(solver,precon);
		break;
/*
	case LIS_MATRIX_CCS:
		lis_matrix_scaling_symm_ccs(A, d);
		break;
	case LIS_MATRIX_MSR:
		lis_matrix_scaling_symm_msr(A, d);
		break;
	case LIS_MATRIX_DIA:
		lis_matrix_scaling_symm_dia(A, d);
		break;
	case LIS_MATRIX_ELL:
		lis_matrix_scaling_symm_ell(A, d);
		break;
	case LIS_MATRIX_JDS:
		lis_matrix_scaling_symm_jds(A, d);
		break;
	case LIS_MATRIX_BSC:
		lis_matrix_scaling_symm_bsc(A, d);
		break;
	case LIS_MATRIX_VBR:
		lis_matrix_scaling_symm_vbr(A, d);
		break;
	case LIS_MATRIX_DNS:
		lis_matrix_scaling_symm_dns(A, d);
		break;
	case LIS_MATRIX_COO:
		lis_matrix_scaling_symm_coo(A, d);
		break;
*/
	default:
		LIS_SETERR_IMP;
		return LIS_ERR_NOT_IMPLEMENTED;
		break;
	}

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_precon_create_ilu0_crs"
int lis_precon_create_ilu0_crs(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				i,j,jj,k,kk,n,np;
	int				jw;
	int				*iw;
	LIS_SCALAR		t,w;
	int				nnzl,nnzu;
	int				err;
	int				*lptr,*lindex,*lg_index,*uptr,*uindex,*ug_index;
	LIS_SCALAR		*lvalue,*uvalue;
	LIS_MATRIX		A,L,U;

	LIS_DEBUG_FUNC_IN;

	A   = solver->A;
	n   = A->n;
	np  = A->np;
	w   = solver->params[LIS_PARAMS_RELAX-LIS_OPTIONS_LEN];

	nnzl     = A->L->nnz;
	nnzu     = A->U->nnz + A->n;
	iw       = NULL;
	lptr     = NULL;
	lindex   = NULL;
	lvalue   = NULL;
	uptr     = NULL;
	uindex   = NULL;
	uvalue   = NULL;

	err = lis_matrix_duplicate(A,&L);
	if( err )
	{
		return err;
	}
	err = lis_matrix_duplicate(A,&U);
	if( err )
	{
		lis_matrix_destroy(L);
		return err;
	}

	iw  = (int *)lis_malloc( np*sizeof(int) );
	if( iw==NULL )
	{
		LIS_SETERR_MEM(np*sizeof(int));
		lis_matrix_destroy(L);
		lis_matrix_destroy(U);
		lis_free2(7,iw,lptr,lindex,lvalue,uptr,uindex,uvalue);
		return LIS_ERR_OUT_OF_MEMORY;
	}

	err = lis_matrix_malloc_crs(n,nnzl,&lptr,&lindex,&lvalue);
	if( err )
	{
		lis_matrix_destroy(L);
		lis_matrix_destroy(U);
		lis_free2(7,iw,lptr,lindex,lvalue,uptr,uindex,uvalue);
		return err;
	}
	err = lis_matrix_malloc_crs(n,nnzu,&uptr,&uindex,&uvalue);
	if( err )
	{
		lis_matrix_destroy(L);
		lis_matrix_destroy(U);
		lis_free2(7,iw,lptr,lindex,lvalue,uptr,uindex,uvalue);
		return err;
	}

	nnzl = 0;
	nnzu = 0;
	lptr[0] = 0;
	uptr[0] = 0;
	for(i=0;i<np;i++) iw[i] = -1;
	for(i=0;i<n;i++)
	{
		nnzl    = lptr[i];
		for(j=A->L->ptr[i];j<A->L->ptr[i+1];j++)
		{
			jj           = A->L->index[j];
			lindex[nnzl] = A->L->index[j];
			lvalue[nnzl] = A->L->value[j];
			iw[jj]       = nnzl;
			nnzl++;
		}
		nnzu    = uptr[i];
		uindex[nnzu] = i;
		uvalue[nnzu] = A->D->value[i];
		iw[i] = nnzu;
		nnzu++;
		for(j=A->U->ptr[i];j<A->U->ptr[i+1];j++)
		{
			jj           = A->U->index[j];
			uindex[nnzu] = A->U->index[j];
			uvalue[nnzu] = A->U->value[j];
			#ifdef USE_MPI
				if( A->U->index[j]>=n )
				{
					continue;
				}
			#endif
			iw[jj]       = nnzu;
			nnzu++;
		}
		lptr[i+1] = nnzl;
		uptr[i+1] = nnzu;

		for(j=lptr[i];j<lptr[i+1];j++)
		{
			jj        = lindex[j];
			/*if( jj>=n ) continue;*/
			t         = lvalue[j] * uvalue[uptr[jj]];
			lvalue[j] = t;

			for(k=uptr[jj]+1;k<uptr[jj+1];k++)
			{
				kk = uindex[k];
				/*if( kk>=n ) continue;*/
				jw = iw[kk];
				if(jw!=-1) uvalue[jw] -= w * t * uvalue[k];
			}
		}
		uvalue[uptr[i]] = 1.0 / uvalue[uptr[i]];

		for(j=lptr[i];j<lptr[i+1];j++) iw[lindex[j]] = -1;
		for(j=uptr[i];j<uptr[i+1];j++) iw[uindex[j]] = -1;
	}

	lis_matrix_set_crs(nnzl,lptr,lindex,lvalue,L);
	lis_matrix_set_crs(nnzu,uptr,uindex,uvalue,U);
	L->is_pmat = LIS_TRUE;
	U->is_pmat = LIS_TRUE;
	lis_matrix_assemble(L);
	lis_matrix_assemble(U);
	lis_free(iw);

	precon->L = L;
	precon->U = U;

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_precon_create_ilu0_bsr"
int lis_precon_create_ilu0_bsr(LIS_SOLVER solver, LIS_PRECON precon)
{
	int				i,j,ii,jj,k,kk,n,gn;
	int				nr,bnr,bnc,bs,bn;
	int				jw;
	int				*iw;
	LIS_SCALAR		t,w;
	int				nnzl,nnzu;
	int				err;
	int				*lptr,*lindex,*lg_index,*uptr,*uindex,*ug_index;
	LIS_SCALAR		*lvalue,*uvalue;
	LIS_MATRIX		A,L,U;

	LIS_DEBUG_FUNC_IN;

	A   = solver->A;
	n   = A->n;
	gn  = A->gn;
	bnr = A->bnr;
	bnc = A->bnc;
	nr  = A->nr;
	bs  = A->bnr*A->bnc;
	w   = solver->params[LIS_PARAMS_RELAX-LIS_OPTIONS_LEN];

	nnzl     = A->L->bnnz*bs;
	nnzu     = A->U->bnnz*bs + n;
	iw       = NULL;
	lptr     = NULL;
	lindex   = NULL;
	lvalue   = NULL;
	uptr     = NULL;
	uindex   = NULL;
	uvalue   = NULL;


	err = lis_matrix_duplicate(A,&L);
	if( err )
	{
		return err;
	}
	err = lis_matrix_duplicate(A,&U);
	if( err )
	{
		lis_matrix_destroy(L);
		return err;
	}

	iw  = (int *)lis_malloc( n*sizeof(int) );
	if( iw==NULL )
	{
		LIS_SETERR_MEM(n*sizeof(int));
		lis_matrix_destroy(L);
		lis_matrix_destroy(U);
		lis_free2(7,iw,lptr,lindex,lvalue,uptr,uindex,uvalue);
		return LIS_ERR_OUT_OF_MEMORY;
	}

	err = lis_matrix_malloc_crs(n,nnzl,&lptr,&lindex,&lvalue);
	if( err )
	{
		lis_matrix_destroy(L);
		lis_matrix_destroy(U);
		lis_free2(7,iw,lptr,lindex,lvalue,uptr,uindex,uvalue);
		return err;
	}
	err = lis_matrix_malloc_crs(n,nnzu,&uptr,&uindex,&uvalue);
	if( err )
	{
		lis_matrix_destroy(L);
		lis_matrix_destroy(U);
		lis_free2(7,iw,lptr,lindex,lvalue,uptr,uindex,uvalue);
		return err;
	}

	nnzl = 0;
	nnzu = 0;
	lptr[0] = 0;
	uptr[0] = 0;
	for(i=0;i<n;i++) iw[i] = -1;
	for(i=0;i<nr;i++)
	{
		for(ii=0;ii<bnr;ii++)
		{
			if( i*bnr+ii==n ) break;
			nnzl    = lptr[i*bnr+ii];
			for(j=A->L->bptr[i];j<A->L->bptr[i+1];j++)
			{
				for(jj=0;jj<bnc;jj++)
				{
					if( A->L->value[j*bs + jj*bnr+ii] != 0.0 )
					{
						k            = A->L->bindex[j]*bnc+jj;
						lindex[nnzl] = A->L->bindex[j]*bnc+jj;
						lvalue[nnzl] = A->L->value[j*bs + jj*bnr+ii];
						iw[k]        = nnzl;
						nnzl++;
					}
				}
			}
			for(j=0;j<ii;j++)
			{
				if( A->D->value[i*bs + j*bnr+ii] != 0.0 )
				{
					lindex[nnzl] = i*bnr+j;
					lvalue[nnzl] = A->D->value[i*bs + j*bnr+ii];
					iw[i*bnr+j]  = nnzl;
					nnzl++;
				}
			}
			nnzu    = uptr[i*bnr+ii];
			for(j=ii;j<bnc;j++)
			{
				if( A->D->value[i*bs + j*bnr+ii] != 0.0 )
				{
					uindex[nnzu] = i*bnr+j;
					uvalue[nnzu] = A->D->value[i*bs + j*bnr+ii];
					iw[i*bnr+j]  = nnzu;
					nnzu++;
				}
			}
			for(j=A->U->bptr[i];j<A->U->bptr[i+1];j++)
			{
				for(jj=0;jj<bnc;jj++)
				{
					if( A->U->value[j*bs + jj*bnr+ii] != 0.0 )
					{
						k            = A->U->bindex[j]*bnc+jj;
						#ifdef USE_MPI
							if( k>=n ) break;
						#endif
						uindex[nnzu] = A->U->bindex[j]*bnc+jj;
						uvalue[nnzu] = A->U->value[j*bs + jj*bnr+ii];
						iw[k]        = nnzu;
						nnzu++;
					}
				}
			}
			lptr[i*bnr+ii+1] = nnzl;
			uptr[i*bnr+ii+1] = nnzu;

			for(j=lptr[i*bnr+ii];j<lptr[i*bnr+ii+1];j++)
			{
				jj        = lindex[j];
				t         = lvalue[j] * uvalue[uptr[jj]];
				lvalue[j] = t;

				for(k=uptr[jj]+1;k<uptr[jj+1];k++)
				{
					kk = uindex[k];
					jw = iw[kk];
					if(jw!=-1) uvalue[jw] -= w * t * uvalue[k];
				}
			}
			uvalue[uptr[i*bnr+ii]] = 1.0 / uvalue[uptr[i*bnr+ii]];

			for(j=lptr[i*bnr+ii];j<lptr[i*bnr+ii+1];j++) iw[lindex[j]] = -1;
			for(j=uptr[i*bnr+ii];j<uptr[i*bnr+ii+1];j++) iw[uindex[j]] = -1;
		}
	}

	lis_matrix_set_crs(nnzl,lptr,lindex,lvalue,L);
	lis_matrix_set_crs(nnzu,uptr,uindex,uvalue,U);
	lis_matrix_assemble(L);
	lis_matrix_assemble(U);
	lis_free(iw);

	precon->L = L;
	precon->U = U;

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolve_ilu0"
int lis_psolve_ilu0(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
#ifdef _OPENMP
	int i,j,n;
	int k,jj,is,ie,threads;
	LIS_SCALAR t,d;
	LIS_SCALAR *b,*x;
	LIS_PRECON precon;

	LIS_DEBUG_FUNC_IN;

	/*
	 *  LUx = b
	 *  LU  = (D + L*A) * (I + D^-1 * U*A)
	 */

	precon = solver->precon;
	n       = precon->L->n;
	b       = B->value;
	x       = X->value;
	threads = omp_get_max_threads();
#pragma omp parallel private(i,j,jj,is,ie,k,t)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);

	for(i=is; i<ie; i++)
	{
		t = b[i];
		for(j=precon->L->ptr[i];j<precon->L->ptr[i+1];j++)
		{
			jj = precon->L->index[j];
			if( jj>=is && jj<ie ) t -= precon->L->value[j] * x[jj];
		}
		x[i] = t;
	}
	for(i=ie-1; i>=is; i--)
	{
		t = x[i];
		d = precon->U->value[precon->U->ptr[i]];
		for(j=precon->U->ptr[i]+1;j<precon->U->ptr[i+1];j++)
		{
			jj = precon->U->index[j];
			if( jj>=is && jj<ie ) t -= precon->U->value[j] * x[jj];
		}
		x[i] = t * d;
	}
}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
#if 0
	int i,j,jj,n,np,k,ads;
	LIS_SCALAR t,d;
	LIS_SCALAR *b,*x;
	LIS_PRECON	precon;

	/*
	 *  LUx = b
	 *  LU  = (D + L*A) * (I + D^-1 * U*A)
	 */

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	n = precon->L->n;
	np = precon->A->np;
	b = B->value;
	x = X->value;
	ads=0;
	for(i=0;i<np;i++)
	{
		x[i] = 0.0;
	}
	for(i=0;i<n;i++)
	{
		precon->temp->value[i] = b[i];
	}
	for(k=0;k<ads+1;k++)
	{
		for(i=n;i<np;i++)
		{
			precon->temp->value[i] = 0.0;
		}
	for(i=0; i<n; i++)
	{
		t = precon->temp->value[i];
		for(j=precon->L->ptr[i];j<precon->L->ptr[i+1];j++)
		{
			jj = precon->L->index[j];
			t -= precon->L->value[j] * precon->temp->value[jj];
		}
		precon->temp->value[i] = t;
	}
	lis_send_recv(precon->A->commtable,precon->temp->value);
	for(i=n-1; i>=0; i--)
	{
		t = precon->temp->value[i];
		d = precon->U->value[precon->U->ptr[i]];
		for(j=precon->U->ptr[i]+1;j<precon->U->ptr[i+1];j++)
		{
			jj = precon->U->index[j];
			t -= precon->U->value[j] * precon->temp->value[jj];
		}
		precon->temp->value[i] = t * d;
	}
	
	for(i=0;i<n;i++)
	{
		x[i] += precon->temp->value[i];
	}
	
	if(k==ads) break;

	lis_send_recv(precon->A->commtable,x);
	
	lis_matvec_crs_xxx[0][0](precon->A,x,precon->temp->value);
	for(i=0;i<n;i++)
	{
		precon->temp->value[i] = b[i] - precon->temp->value[i];
	}
	
	/*
	for(i=0;i<n;i++)
	{
		t = b[i] - x[i]/precon->U->value[precon->U->ptr[i]];
		for(j=precon->L->ptr[i];j<precon->L->ptr[i+1];j++)
		{
			jj = precon->L->index[j];
			t -= precon->L->value[j] * x[jj];
		}
		for(j=precon->U->ptr[i]+1;j<precon->U->ptr[i+1];j++)
		{
			jj = precon->U->index[j];
			t -= precon->U->value[j] * x[jj];
		}
		precon->temp->value[i] = t;
	}
	*/
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
	int i,j,jj,n;
	LIS_SCALAR t,d;
	LIS_SCALAR *b,*x;
	LIS_PRECON	precon;

	/*
	 *  LUx = b
	 *  LU  = (D + L*A) * (I + D^-1 * U*A)
	 */

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	n = precon->L->n;
	b = B->value;
	x = X->value;
	for(i=0; i<n; i++)
	{
		t = b[i];
		for(j=precon->L->ptr[i];j<precon->L->ptr[i+1];j++)
		{
			jj = precon->L->index[j];
			t -= precon->L->value[j] * x[jj];
		}
		x[i] = t;
	}
	for(i=n-1; i>=0; i--)
	{
		t = x[i];
		d = precon->U->value[precon->U->ptr[i]];
		for(j=precon->U->ptr[i]+1;j<precon->U->ptr[i+1];j++)
		{
			jj = precon->U->index[j];
			t -= precon->U->value[j] * x[jj];
		}
		x[i] = t * d;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}

#undef __FUNC__
#define __FUNC__ "lis_psolvet_ilu0"
int lis_psolvet_ilu0(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
#ifdef _OPENMP
	int i,j,n;
	int k,jj,is,ie,threads;
	LIS_SCALAR t,d;
	LIS_SCALAR *b,*x;
	LIS_PRECON precon;

	LIS_DEBUG_FUNC_IN;

	/*
	 *  (LU)'x = b
	 *  LU     = (D + L*A) * (I + D^-1 * U*A)
	 */

	precon = solver->precon;
	n = precon->L->n;
	b = B->value;
	x = X->value;
	threads = omp_get_max_threads();
#pragma omp parallel private(i,j,jj,is,ie,k,t)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);

	for(i=is; i<ie; i++)
	{
		x[i] = b[i];
	}
	for(i=is; i<ie; i++)
	{
		x[i] = x[i] * precon->U->value[precon->U->ptr[i]];
		for(j=precon->U->ptr[i]+1;j<precon->U->ptr[i+1];j++)
		{
			jj     = precon->U->index[j];
			if( jj>=is && jj<ie ) x[jj] -= precon->U->value[j] * x[i];
		}
	}
	for(i=ie-1; i>=is; i--)
	{
		for(j=precon->L->ptr[i];j<precon->L->ptr[i+1];j++)
		{
			jj     = precon->L->index[j];
			if( jj>=is && jj<ie ) x[jj] -= precon->L->value[j] * x[i];
		}
	}
}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#else
	int i,j,jj,n;
	LIS_SCALAR t,d;
	LIS_SCALAR *b,*x;
	LIS_PRECON	precon;

	/*
	 *  (LU)'x = b
	 *  U'x=b
	 *  L'x=b
	 */

	LIS_DEBUG_FUNC_IN;

	precon = solver->precon;
	n = precon->L->n;
	b = B->value;
	x = X->value;
	lis_vector_copy(B,X);
	for(i=0; i<n; i++)
	{
		x[i] = x[i] * precon->U->value[precon->U->ptr[i]];
		for(j=precon->U->ptr[i]+1;j<precon->U->ptr[i+1];j++)
		{
			jj     = precon->U->index[j];
			x[jj] -= precon->U->value[j] * x[i];
		}
	}
	for(i=n-1; i>=0; i--)
	{
		for(j=precon->L->ptr[i];j<precon->L->ptr[i+1];j++)
		{
			jj     = precon->L->index[j];
			x[jj] -= precon->L->value[j] * x[i];
		}
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}

#if 0
#undef __FUNC__
#define __FUNC__ "lis_precon_post_ilu0"
int lis_precon_post_ilu0(LIS_PRECON precon, LIS_VECTOR B, LIS_VECTOR X)
{
#ifdef _OPENMP
	int i,j,n;
	int k,jj,is,ie,threads;
	LIS_SCALAR t;
	LIS_SCALAR *b,*x;

	LIS_DEBUG_FUNC_IN;

	/*
	 *  LUx = b
	 *  LU  = (D + L*A) * (I + D^-1 * U*A)
	 */

	n       = precon->Pmat->n;
	b       = B->value;
	x       = X->value;
	threads = omp_get_max_threads();
#pragma omp parallel private(i,j,jj,is,ie,k,t)
{
	k = omp_get_thread_num();
	LIS_GET_ISIE(k,threads,n,is,ie);

	for(i=is; i<ie; i++)
	{
		t = b[i];
		for(j=precon->Pmat->index[i];j<precon->Pmat->ptr[i];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) t -= precon->Pmat->value[j] * x[jj];
		}
		x[i] = t;
	}
	for(i=ie-1; i>=is; i--)
	{
		t = x[i];
		for(j=precon->Pmat->ptr[i];j<precon->Pmat->index[i+1];j++)
		{
			jj = precon->Pmat->index[j];
			if( jj>=is && jj<ie ) t -= precon->Pmat->value[j] * x[jj];
		}
		x[i] = t * precon->Pmat->value[i];
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#else
	int i,j,jj,n;
	LIS_SCALAR t,d;
	LIS_SCALAR *b,*x;

	/*
	 *  x = LUb
	 */

	LIS_DEBUG_FUNC_IN;

	n = precon->L->n;
	b = B->value;
	x = X->value;
	for(i=0; i<n; i++)
	{
		t = 1.0 / precon->U->value[precon->U->ptr[i]];
		t = t * b[i];
		for(j=precon->U->ptr[i]+1;j<precon->U->ptr[i+1];j++)
		{
			jj = precon->U->index[j];
			t += precon->U->value[j] * b[jj];
		}
		x[i] = t;
	}
	for(i=0; i<n; i++)
	{
		t = x[i];
		for(j=precon->L->ptr[i];j<precon->L->ptr[i+1];j++)
		{
			jj = precon->L->index[j];
			t += precon->L->value[j] * x[jj];
		}
		b[i] = t;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
#endif
}
#endif
