/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef USE_SSE2
	#include <emmintrin.h>
#endif
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_precon_create
 * lis_psolve
 * lis_psolvet
 ************************************************/

#undef __FUNC__
#define __FUNC__ "lis_precon_create_jacobi"
int lis_precon_create_jacobi(LIS_SOLVER solver, LIS_PRECON precon)
{
	int			i,n,err;
	LIS_SCALAR	*D;

	LIS_DEBUG_FUNC_IN;

	if( solver->precision==LIS_PRECISION_DEFAULT )
	{
		err = lis_vector_duplicate(solver->A, &precon->D);
	}
	else
	{
		err = lis_vector_duplicateex(LIS_PRECISION_QUAD,solver->A, &precon->D);
	}
	if( err )
	{
		return err;
	}

	lis_matrix_get_diagonal(solver->A, precon->D);
	if( solver->precision==LIS_PRECISION_DEFAULT )
	{
		lis_vector_reciprocal(precon->D);
	}
#ifdef USE_QUAD_PRECISION
	else
	{
		lis_vector_reciprocalex_m(precon->D);
	}
#endif

	LIS_DEBUG_FUNC_OUT;
    return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolve_jacobi"
int lis_psolve_jacobi(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,n;
	LIS_SCALAR *b,*x,*d;
	LIS_SCALAR *bl,*xl,*dl;
	LIS_PRECON	precon;
			#ifndef USE_SSE2
				double p1, p2;
				double t,bhi,blo,chi,clo;
			#else
				__m128d bch,bcl,xs,xt,xr,xw,xe,xb;
			#endif
	LIS_DEBUG_FUNC_IN;

	/*
	 *  Mx = b
	 *  M  = D
	 */

	precon = solver->precon;
	n = precon->D->n;
	d = precon->D->value;
	b = B->value;
	x = X->value;
	dl = precon->D->value_lo;
	bl = B->value_lo;
	xl = X->value_lo;

	if( B->precision==LIS_PRECISION_DEFAULT )
	{
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			x[i] = b[i] * d[i];
		}
	}
	else
	{
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
/*			x[i] = b[i] * d[i];*/
			#ifndef USE_SSE2

				p1 =  b[i] * d[i];
				t   = SPLITTER * b[i];
				bhi = t - (t-b[i]);
				blo = b[i] - bhi;
				t   = SPLITTER * d[i];
				chi = t - (t-d[i]);
				clo = d[i] - chi;
				p2 = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo;
				p2 += (b[i] * dl[i]);
				p2 += (bl[i] * d[i]);
				x[i] = p1 + p2;
				xl[i] = p2  - (x[i] - p1);
			#else

				xs  = _mm_set_pd(SPLITTER,SPLITTER);
				bch = _mm_set_pd(b[i],d[i]);
				xb  = _mm_unpackhi_pd(bch,bch); /* bh,bh */
				bcl = _mm_set_pd(dl[i],bl[i]);
				xb  = _mm_mul_sd(xb,bch);
				xs  = _mm_mul_pd(xs,bch);
				xt  = _mm_sub_pd(xs,bch);
				xs  = _mm_sub_pd(xs,xt);
				xt  = _mm_mul_pd(bch,bcl);
				bch = _mm_sub_pd(bch,xs);
				xe  = _mm_unpackhi_pd(xs,bch);  /* blo,sh */
				xw  = _mm_unpacklo_pd(xs,bch);  /* clo,sl */
				bch = _mm_unpackhi_pd(bch,xw);  /* clo,blo */
				xw  = _mm_mul_pd(xw,xe);
				xs  = _mm_mul_pd(xs,bch);
				bch = _mm_unpackhi_pd(xs,xs);
				bcl = _mm_unpackhi_pd(xw,xw);
				xe  = _mm_unpackhi_pd(xt,xt);
				xw  = _mm_sub_sd(xw,xb);
				xw  = _mm_add_sd(xw,bch);
				xw  = _mm_add_sd(xw,xs);
				xw  = _mm_add_sd(xw,bcl);
				xw  = _mm_add_sd(xw,xe);
				xw  = _mm_add_sd(xw,xt);
				xt  = _mm_add_sd(xw,xb);
				_mm_store_sd(&x[i],xt);
				xt  = _mm_sub_sd(xt,xb);
				xw  = _mm_sub_sd(xw,xt);
				_mm_store_sd(&xl[i],xw);
			#endif
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_psolvet_jacobi"
int lis_psolvet_jacobi(LIS_SOLVER solver, LIS_VECTOR B, LIS_VECTOR X)
{
	int i,n;
	LIS_SCALAR *b,*x,*d;
	LIS_SCALAR *bl,*xl,*dl;
	LIS_PRECON	precon;
			#ifndef USE_SSE2
				double p1, p2;
				double t,bhi,blo,chi,clo;
			#else
				__m128d bch,bcl,xs,xt,xr,xw,xe,xb;
			#endif
	LIS_DEBUG_FUNC_IN;

	/*
	 *  Mx = b
	 *  M  = D
	 */

	precon = solver->precon;
	n = precon->D->n;
	d = precon->D->value;
	b = B->value;
	x = X->value;
	dl = precon->D->value_lo;
	bl = B->value_lo;
	xl = X->value_lo;

	if( B->precision==LIS_PRECISION_DEFAULT )
	{
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			x[i] = b[i] * d[i];
		}
	}
	else
	{
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
/*			x[i] = b[i] * d[i];*/
			#ifndef USE_SSE2

				p1 =  b[i] * d[i];
				t   = SPLITTER * b[i];
				bhi = t - (t-b[i]);
				blo = b[i] - bhi;
				t   = SPLITTER * d[i];
				chi = t - (t-d[i]);
				clo = d[i] - chi;
				p2 = ((bhi*chi-p1)+bhi*clo+blo*chi)+blo*clo;
				p2 += (b[i] * dl[i]);
				p2 += (bl[i] * d[i]);
				x[i] = p1 + p2;
				xl[i] = p2  - (x[i] - p1);
			#else

				xs  = _mm_set_pd(SPLITTER,SPLITTER);
				bch = _mm_set_pd(b[i],d[i]);
				xb  = _mm_unpackhi_pd(bch,bch); /* bh,bh */
				bcl = _mm_set_pd(dl[i],bl[i]);
				xb  = _mm_mul_sd(xb,bch);
				xs  = _mm_mul_pd(xs,bch);
				xt  = _mm_sub_pd(xs,bch);
				xs  = _mm_sub_pd(xs,xt);
				xt  = _mm_mul_pd(bch,bcl);
				bch = _mm_sub_pd(bch,xs);
				xe  = _mm_unpackhi_pd(xs,bch);  /* blo,sh */
				xw  = _mm_unpacklo_pd(xs,bch);  /* clo,sl */
				bch = _mm_unpackhi_pd(bch,xw);  /* clo,blo */
				xw  = _mm_mul_pd(xw,xe);
				xs  = _mm_mul_pd(xs,bch);
				bch = _mm_unpackhi_pd(xs,xs);
				bcl = _mm_unpackhi_pd(xw,xw);
				xe  = _mm_unpackhi_pd(xt,xt);
				xw  = _mm_sub_sd(xw,xb);
				xw  = _mm_add_sd(xw,bch);
				xw  = _mm_add_sd(xw,xs);
				xw  = _mm_add_sd(xw,bcl);
				xw  = _mm_add_sd(xw,xe);
				xw  = _mm_add_sd(xw,xt);
				xt  = _mm_add_sd(xw,xb);
				_mm_store_sd(&x[i],xt);
				xt  = _mm_sub_sd(xt,xb);
				xw  = _mm_sub_sd(xw,xt);
				_mm_store_sd(&xl[i],xw);
			#endif
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
