/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_solver_init
 * lis_solver_create
 * lis_solver_destroy
 * lis_solver_work_destroy
 * lis_solver_set_option
 * lis_solver_get_option
 * lis_solve
 ************************************************/

#define LIS_SOLVERS_LEN			12
#define LIS_PRECON_TYPE_LEN		10


LIS_SOLVER_CHECK_PARAMS lis_solver_check_params[] = {
	NULL,
	lis_cg_check_params       , lis_bicg_check_params      , lis_cgs_check_params, 
	lis_bicgstab_check_params , lis_bicgstabl_check_params , lis_gpbicg_check_params, 
	lis_orthomin_check_params , lis_tfqmr_check_params     , lis_gmres_check_params,
	lis_jacobi_check_params   , lis_gs_check_params        , lis_sor_check_params
};

LIS_SOLVER_MALLOC_WORK lis_solver_malloc_work[] = {
	NULL,
	lis_cg_malloc_work       , lis_bicg_malloc_work      , lis_cgs_malloc_work, 
	lis_bicgstab_malloc_work , lis_bicgstabl_malloc_work , lis_gpbicg_malloc_work, 
	lis_orthomin_malloc_work , lis_tfqmr_malloc_work     , lis_gmres_malloc_work,
	lis_jacobi_malloc_work   , lis_gs_malloc_work        , lis_sor_malloc_work
};

LIS_SOLVER_EXECUTE lis_solver_execute[] = {
	NULL,
	lis_cg       , lis_bicg      , lis_cgs, 
	lis_bicgstab , lis_bicgstabl , lis_gpbicg, 
	lis_orthomin , lis_tfqmr     , lis_gmres,
	lis_jacobi   , lis_gs        , lis_sor
};

#ifdef USE_QUAD_PRECISION
	LIS_SOLVER_EXECUTE lis_solver_execute_quad[] = {
		NULL,
		lis_cg_quad       , lis_bicg_quad  , lis_cgs_quad, 
		lis_bicgstab_quad , NULL           , lis_gpbicg_quad,
		NULL              , lis_tfqmr_quad , lis_gmres_quad,
		NULL              , NULL           , NULL,
	};
	LIS_SOLVER_EXECUTE lis_solver_execute_switch[] = {
		NULL,
		lis_cg_switch       , lis_bicg_switch , lis_cgs_switch, 
		lis_bicgstab_switch , NULL            , lis_gpbicg_switch,
		NULL                , NULL            , NULL,
		NULL                , NULL            , NULL,
	};
	LIS_SOLVER_EXECUTE lis_solver_execute_periodic[] = {
		NULL,
		lis_cg_periodic       , lis_bicg_periodic , lis_cgs_periodic, 
		lis_bicgstab_periodic , NULL              , lis_gpbicg_periodic,
		NULL                  , NULL              , NULL,
		NULL                  , NULL              , NULL,
	};
#endif


int LIS_USE_AT_TYPE[] = {
	0,
	LIS_MATRIX_CCS,LIS_MATRIX_CRS
	};
#define LIS_SOLVER_OPTION_LEN	39
#define LIS_PRINT_LEN			4
#define LIS_SCALE_LEN			3
#define LIS_TRUEFALSE_LEN		2
#define LIS_PRECISION_LEN		4

char *LIS_SOLVER_OPTNAME[] = {
	"-maxiter",           "-tol",        "-print",          "-scale",       "-ssor_w",
	"-ilu_fill",          "-ilu_relax",  "-is_alpha",       "-is_level",    "-is_m",
	"-hybrid_maxiter",    "-hybrid_ell", "-hybrid_restart", "-hybrid_tol",  "-hybrid_w",
	"-hybrid_i",          "-sainv_drop", "-ric2s_tau",      "-ric2s_sigma", "-ric2s_gamma",
	"-restart",           "-ell",        "-omega",          "-i",           "-p",
	"-f",                 "-h",          "-ver",            "-hybrid_p",    "-initx_zeros",
	"-adds",              "-adds_iter",  "-precision",      "-use_at",      "-periodic_num",
	"-periodic_interval", "-switch_tol", "-switch_maxiter", "-saamg_unsym"
};

int LIS_SOLVER_OPTACT[] = {
	LIS_OPTIONS_MAXITER          , LIS_PARAMS_RESID       , LIS_OPTIONS_OUTPUT        , LIS_OPTIONS_SCALE       , LIS_PARAMS_SSOR_W ,
	LIS_OPTIONS_FILL             , LIS_PARAMS_RELAX       , LIS_PARAMS_ALPHA          , LIS_OPTIONS_ISLEVEL     , LIS_OPTIONS_M     ,
	LIS_OPTIONS_PMAXITER         , LIS_OPTIONS_PELL       , LIS_OPTIONS_PRESTART      , LIS_PARAMS_PRESID       , LIS_PARAMS_PW     ,
	LIS_OPTIONS_PSOLVER          , LIS_PARAMS_DROP        , LIS_PARAMS_TAU            , LIS_PARAMS_SIGMA        , LIS_PARAMS_GAMMA  ,
	LIS_OPTIONS_RESTART          , LIS_OPTIONS_ELL        , LIS_PARAMS_W              , LIS_OPTIONS_SOLVER      , LIS_OPTIONS_PRECON,
	LIS_OPTIONS_FILE             , LIS_OPTIONS_HELP       , LIS_OPTIONS_VER           , LIS_OPTIONS_PPRECON     , LIS_OPTIONS_INITGUESS_ZEROS,
	LIS_OPTIONS_ADDS             , LIS_OPTIONS_ADDS_ITER  , LIS_OPTIONS_PRECISION     , LIS_OPTIONS_USE_AT      , LIS_OPTIONS_PERIODIC_NUM,
	LIS_OPTIONS_PERIODIC_INTERVAL, LIS_PARAMS_SWITCH_RESID, LIS_OPTIONS_SWITCH_MAXITER, LIS_OPTIONS_SAAMG_UNSYM
};

char *lis_solver_atoi[]    = {"cg", "bicg", "cgs", "bicgstab", "bicgstabl", "gpbicg", "orthomin", "tfqmr", "gmres", "jacobi", "gs", "sor"};
char *lis_precon_atoi[]    = {"none", "jacobi", "ilu", "ssor", "hybrid", "is", "sainv", "saamg", "iluc"};
char *lis_storage_atoi[]   = {"csr", "ccs", "msr", "dia", "ell", "jds", "bsr", "bsc", "vbr", "coo", "dns"};
char *lis_print_atoi[]     = {"none", "mem", "out", "all"};
char *lis_scale_atoi[]     = {"none", "jacobi", "symm_diag"};
char *lis_truefalse_atoi[] = {"false", "true"};
char *lis_precision_atoi[] = {"double", "quad", "switch", "periodic"};



char *lis_solvername[] = {"", "CG", "BiCG", "CGS", "BiCGSTAB", "BiCGSTAB(l)", "GPBICG", 
						"orthomin", "TFQMR", "gmres", "Jacobi",	"Gauss-Seidel", "SOR"};
char *lis_preconname[] = {"None", "Jacobi", "ILU", "SSOR", "HYBRID", "I+S", "SAINV", "SAAMG", "Crout ILU"};
char *lis_returncode[] = {"LIS_SUCCESS", "LIS_ILL_OPTION", "LIS_BREAKDOWN", "LIS_OUT_OF_MEMORY", "LIS_MAXITER", "LIS_NOT_IMPLEMENTED", "LIS_ERR_FILE_IO"};

LIS_VECTOR lis_residual_history = NULL;
int lis_residual_history_count = 0;

#undef __FUNC__
#define __FUNC__ "lis_solver_init"
int lis_solver_init(LIS_SOLVER solver)
{

	LIS_DEBUG_FUNC_IN;

	solver->A        = NULL;
	solver->At       = NULL;
	solver->b        = NULL;
	solver->x        = NULL;
	solver->work     = NULL;
	solver->residual = NULL;
	solver->precon   = NULL;

	solver->worklen  = 0;
	solver->iter     = 0;
	solver->resid    = 0;
	solver->times    = 0;
	solver->itimes   = 0;
	solver->ptimes   = 0;

	solver->options[LIS_OPTIONS_SOLVER]               = LIS_SOLVER_BICG;
	solver->options[LIS_OPTIONS_PRECON]               = LIS_PRECON_TYPE_NONE;
	solver->options[LIS_OPTIONS_OUTPUT]               = LIS_FALSE;
	solver->options[LIS_OPTIONS_MAXITER]              = 1000;
	solver->options[LIS_OPTIONS_RESTART]              = 40;
	solver->options[LIS_OPTIONS_ELL]                  = 2;
	solver->options[LIS_OPTIONS_SCALE]                = LIS_SCALE_NONE;
	solver->options[LIS_OPTIONS_FILL]                 = 0;
	solver->options[LIS_OPTIONS_M]                    = 3;
	solver->options[LIS_OPTIONS_PSOLVER]              = LIS_SOLVER_SOR;
	solver->options[LIS_OPTIONS_PMAXITER]             = 10;
	solver->options[LIS_OPTIONS_PRESTART]             = 40;
	solver->options[LIS_OPTIONS_PELL]                 = 2;
	solver->options[LIS_OPTIONS_PPRECON]              = LIS_PRECON_TYPE_NONE;
	solver->options[LIS_OPTIONS_ISLEVEL]              = 1;
	solver->options[LIS_OPTIONS_INITGUESS_ZEROS]      = LIS_TRUE;
	solver->options[LIS_OPTIONS_ADDS]                 = LIS_FALSE;
	solver->options[LIS_OPTIONS_ADDS_ITER]            = 1;
	solver->options[LIS_OPTIONS_PRECISION]            = LIS_PRECISION_DOUBLE;
	solver->options[LIS_OPTIONS_PERIODIC_INTERVAL]    = 10;
	solver->options[LIS_OPTIONS_PERIODIC_NUM]         = 1;
	solver->options[LIS_OPTIONS_USE_AT]               = LIS_FALSE;
	solver->options[LIS_OPTIONS_SWITCH_MAXITER]       = -1;
	solver->options[LIS_OPTIONS_SAAMG_UNSYM]          = LIS_FALSE;

	solver->params[LIS_PARAMS_RESID        -LIS_OPTIONS_LEN] = 1.0e-12;
	solver->params[LIS_PARAMS_W            -LIS_OPTIONS_LEN] = 1.9;
	solver->params[LIS_PARAMS_SSOR_W       -LIS_OPTIONS_LEN] = 1.0;
	solver->params[LIS_PARAMS_RELAX        -LIS_OPTIONS_LEN] = 1.0;
	solver->params[LIS_PARAMS_DROP         -LIS_OPTIONS_LEN] = 0.05;
	solver->params[LIS_PARAMS_ALPHA        -LIS_OPTIONS_LEN] = 1.0;
	solver->params[LIS_PARAMS_TAU          -LIS_OPTIONS_LEN] = 0.05;
	solver->params[LIS_PARAMS_SIGMA        -LIS_OPTIONS_LEN] = 2.0;
	solver->params[LIS_PARAMS_GAMMA        -LIS_OPTIONS_LEN] = 1.0;
	solver->params[LIS_PARAMS_PRESID       -LIS_OPTIONS_LEN] = 1.0e-2;
	solver->params[LIS_PARAMS_PW           -LIS_OPTIONS_LEN] = 1.5;
	solver->params[LIS_PARAMS_SWITCH_RESID -LIS_OPTIONS_LEN] = 1.0e-12;
	solver->params[LIS_PARAMS_RATE         -LIS_OPTIONS_LEN] = 5.0;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_create"
int lis_solver_create(LIS_SOLVER *solver)
{
	LIS_DEBUG_FUNC_IN;

	*solver = NULL;

	*solver = (LIS_SOLVER)lis_malloc( sizeof(struct LIS_SOLVER_STRUCT) );
	if( NULL==*solver )
	{
		LIS_SETERR_MEM(sizeof(struct LIS_SOLVER_STRUCT));
		return LIS_OUT_OF_MEMORY;
	}
	lis_solver_init(*solver);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_work_destroy"
int lis_solver_work_destroy(LIS_SOLVER solver)
{
	int i;

	LIS_DEBUG_FUNC_IN;

	if( solver )
	{
		for(i=0;i<solver->worklen;i++) lis_vector_destroy(solver->work[i]);
		lis_free(solver->work);
		solver->work    = NULL;
		solver->worklen = 0;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_destroy"
int lis_solver_destroy(LIS_SOLVER solver)
{
	LIS_DEBUG_FUNC_IN;

	if( solver )
	{
		lis_free(solver);
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_solve"
int lis_solve(LIS_MATRIX A, LIS_VECTOR b, LIS_VECTOR x, LIS_SOLVER solver)
{
	int			nsolver, precon_type, maxiter;
	int			err;
	LIS_SCALAR	*residual;
	LIS_VECTOR	xx;

	int restart;
	int output;
	int scale;
	int precision,is_use_at;
	int i,n,np;
	int worklen;
	double ptimes, itimes;
	LIS_SCALAR nrm2;
	LIS_VECTOR d,t;
	LIS_VECTOR *work;
	LIS_VECTOR bb;
	int *perm;
	LIS_PRECON precon;
	LIS_MATRIX AA;
	LIS_MATRIX At;

	LIS_DEBUG_FUNC_IN;

	/* begin parameter check */
	err = lis_matrix_check(A,LIS_MATRIX_CHECK_ALL);
	if( err ) return err;
	if( b==NULL )
	{
		LIS_SETERR(LIS_ERR_ILL_ARG,"vector b is undefined\n");
		return LIS_ERR_ILL_ARG;
	}
	if( x==NULL )
	{
		LIS_SETERR(LIS_ERR_ILL_ARG,"vector x is undefined\n");
		return LIS_ERR_ILL_ARG;
	}
	if( A->n!=b->n || A->n!=x->n )
	{
		LIS_SETERR3(LIS_ERR_ILL_ARG,"dimension different (A=%d b=%d x=%d)\n",A->n,b->n,x->n);
		return LIS_ERR_ILL_ARG;
	}
	if( A->n<=0 )
	{
		LIS_SETERR1(LIS_ERR_ILL_ARG,"Dimension n(=%d) of matrix A is less than 0\n",A->n);
		return LIS_ERR_ILL_ARG;
	}

	nsolver     = solver->options[LIS_OPTIONS_SOLVER];
	precon_type = solver->options[LIS_OPTIONS_PRECON];
	maxiter     = solver->options[LIS_OPTIONS_MAXITER];
	output      = solver->options[LIS_OPTIONS_OUTPUT];
	scale       = solver->options[LIS_OPTIONS_SCALE];
	precision   = solver->options[LIS_OPTIONS_PRECISION];
	is_use_at   = solver->options[LIS_OPTIONS_USE_AT];
	solver->precision = precision;

	if( nsolver < 1 || nsolver > LIS_SOLVERS_LEN )
	{
		LIS_SETERR2(LIS_ERR_ILL_ARG,"Parameter LIS_OPTIONS_SOLVER is %d (Set between 1 to %d)\n",nsolver, LIS_SOLVERS_LEN);
		return LIS_ERR_ILL_ARG;
	}
	if( precon_type < 0 || precon_type > precon_register_type )
	{
		LIS_SETERR2(LIS_ERR_ILL_ARG,"Parameter LIS_OPTIONS_PRECON is %d (Set between 0 to %d)\n",precon_type, precon_register_type-1);
		return LIS_ERR_ILL_ARG;
	}
	if( maxiter<0 )
	{
		LIS_SETERR1(LIS_ERR_ILL_ARG,"Parameter LIS_OPTIONS_MAXITER(=%d) is less than 0\n",maxiter);
		return LIS_ERR_ILL_ARG;
	}
	#ifdef USE_QUAD_PRECISION
		if( precision==LIS_PRECISION_QUAD && lis_solver_execute_quad[nsolver]==NULL )
		{
			LIS_SETERR1(LIS_ERR_NOT_IMPLEMENTED,"Quad precision solver %s is not implemented\n",lis_solvername[nsolver]);
			return LIS_ERR_NOT_IMPLEMENTED;
		}
		else if( precision==LIS_PRECISION_SWITCH && lis_solver_execute_switch[nsolver]==NULL )
		{
			LIS_SETERR1(LIS_ERR_NOT_IMPLEMENTED,"Switch solver %s is not implemented\n",lis_solvername[nsolver]);
			return LIS_ERR_NOT_IMPLEMENTED;
		}
		else if( precision==LIS_PRECISION_PERIODIC && lis_solver_execute_periodic[nsolver]==NULL )
		{
			LIS_SETERR1(LIS_ERR_NOT_IMPLEMENTED,"Periodic solver %s is not implemented\n",lis_solvername[nsolver]);
			return LIS_ERR_NOT_IMPLEMENTED;
		}
		if( solver->options[LIS_OPTIONS_SWITCH_MAXITER]==-1 )
		{
			solver->options[LIS_OPTIONS_SWITCH_MAXITER] = maxiter;
		}
	#endif

	err = lis_solver_check_params[nsolver](solver);
	if( err )
	{
		solver->retcode = err;
		return err;
	}
	/* end parameter check */

	solver->A        = A;
	solver->b        = b;

	/* create initial vector */
	#ifndef USE_QUAD_PRECISION
		err = lis_vector_duplicate(A,&xx);
	#else
		if( precision==LIS_PRECISION_DOUBLE )
		{
			err = lis_vector_duplicate(A,&xx);
		}
		else
		{
			err = lis_vector_duplicateex(LIS_PRECISION_QUAD,A,&xx);
		}
	#endif
	if( err )
	{
		solver->retcode = err;
		return err;
	}
	if( solver->options[LIS_OPTIONS_INITGUESS_ZEROS] )
	{
		lis_printf(A->comm,"Initial vector x = 0\n");
		#ifndef USE_QUAD_PRECISION
			lis_vector_set_all(0.0,xx);
		#else
			if( precision==LIS_PRECISION_DOUBLE )
			{
				lis_vector_set_all(0.0,xx);
			}
			else
			{
				lis_vector_set_allex_nm(0.0,xx);
			}
		#endif
	}
	else
	{
		lis_printf(A->comm,"Initial vector x = user defined\n");
		#ifndef USE_QUAD_PRECISION
			lis_vector_copy(x,xx);
		#else
			if( precision==LIS_PRECISION_DOUBLE )
			{
				lis_vector_copy(x,xx);
			}
			else
			{
				lis_vector_copyex_nm(x,xx);
			}
		#endif
	}

	/* create residual history vector */
	residual = lis_malloc((maxiter+1)*sizeof(LIS_SCALAR));
	if( residual==NULL )
	{
		LIS_SETERR_MEM((maxiter+1)*sizeof(LIS_SCALAR));
		lis_vector_destroy(xx);
		solver->retcode = err;
		return err;
	}


	n       = A->n;
	np      = A->np;
	d       = NULL;
	t       = NULL;
	At      = NULL;

	if( A->my_rank==0 )
	{
		printf("SOLVER : %s %d\n", lis_solvername[nsolver],nsolver);
		if( solver->options[LIS_OPTIONS_ADDS] && precon_type )
		{
			printf("PRECON : %s + additive schwarz\n", lis_preconname[precon_type]);
		}
		else
		{
			printf("PRECON : %s\n", lis_preconname[precon_type]);
		}
	}

	ptimes = lis_wtime();
	if( precon_type==LIS_PRECON_TYPE_IS )
	{
		err = lis_vector_duplicate(x,&d);
		if( err )
		{
			return err;
		}
		lis_matrix_scaling(A,b,d,LIS_SCALE_JACOBI);
		if( nsolver >= LIS_SOLVER_JACOBI )
		{
			solver->options[LIS_OPTIONS_ISLEVEL] = 0;
		}
	}
	else if( scale )
	{
		err = lis_vector_duplicate(x,&d);
		if( err )
		{
			return err;
		}
		if( scale==LIS_SCALE_JACOBI && nsolver==LIS_SOLVER_CG )
		{
			scale = LIS_SCALE_SYMM_DIAG;
		}
		lis_matrix_scaling(A,b,d,scale);
	}

	/* create preconditioner */
	err = lis_precon_create(solver, &precon);
	if( err )
	{
		lis_vector_destroy(xx);
		lis_solver_work_destroy(solver);
		lis_free(residual);
		solver->retcode = err;
		return err;
	}
	if( precon_type==LIS_PRECON_TYPE_IS )
	{
		if( nsolver < LIS_SOLVER_JACOBI )
		{
			AA = A;
			bb = b;
		}
		else
		{
			AA = precon->A;
			bb = precon->Pb;
		}
	}
	else
	{
		AA = A;
		bb = b;
	}

	ptimes = lis_wtime() - ptimes;
	itimes = lis_wtime();

	solver->A  = AA;
	solver->b  = bb;

	/* create work vector */
	err = lis_solver_malloc_work[nsolver](solver);
	if( err )
	{
		lis_vector_destroy(d);
		lis_vector_destroy(xx);
		lis_precon_destroy(precon);
		solver->retcode = err;
		return err;
	}
	if( nsolver==LIS_SOLVER_BICG && is_use_at )
	{
		lis_printf(A->comm,"Use At\n");
		lis_matrix_duplicate(AA,&At);
		lis_matrix_set_type(At,LIS_USE_AT_TYPE[AA->matrix_type]);
		lis_matrix_convert(AA,At);
		solver->At = At;
	}

	solver->x        = xx;
	solver->xx       = x;
	solver->precon   = precon;
	solver->residual = residual;

	/* execute solver */
	#ifndef USE_QUAD_PRECISION
		err = lis_solver_execute[nsolver](solver);
	#else
		if( precision==LIS_PRECISION_DOUBLE )
		{
			err = lis_solver_execute[nsolver](solver);
		}
		else if( precision==LIS_PRECISION_QUAD )
		{
			err = lis_solver_execute_quad[nsolver](solver);
		}
		else if( precision==LIS_PRECISION_SWITCH )
		{
			err = lis_solver_execute_switch[nsolver](solver);
		}
		else if( precision==LIS_PRECISION_PERIODIC )
		{
			err = lis_solver_execute_periodic[nsolver](solver);
		}
	#endif
	solver->retcode = err;

	if( scale==LIS_SCALE_SYMM_DIAG && precon_type!=LIS_PRECON_TYPE_IS)
	{
		#ifndef USE_QUAD_PRECISION
			#pragma omp parallel for
			for(i=0;i<n;i++)
			{
				x->value[i] = xx->value[i]*d->value[i];
			}
		#else
			if( precision==LIS_PRECISION_DOUBLE )
			{
				#pragma omp parallel for
				for(i=0;i<n;i++)
				{
					x->value[i] = xx->value[i]*d->value[i];
				}
			}
			else
			{
				#pragma omp parallel for
				for(i=0;i<n;i++)
				{
					x->value[i] = ((LIS_MSCALAR *)xx->value)[i].hi*d->value[i];
				}
			}
		#endif
	}
	else
	{
		#ifndef USE_QUAD_PRECISION
			lis_vector_copy(xx,x);
		#else
			if( precision==LIS_PRECISION_DOUBLE )
			{
				lis_vector_copy(xx,x);
			}
			else
			{
				lis_vector_copyex_mn(xx,x);
			}
		#endif
	}
	itimes = lis_wtime() - itimes - solver->ptimes;
	ptimes += solver->ptimes;
	solver->times  = ptimes + itimes;
	solver->ptimes = ptimes;
	solver->itimes = itimes;
	lis_solver_work_destroy(solver);
	lis_vector_duplicate(A,&t);
	lis_matvec(A,x,t);
	lis_vector_xpay(b,-1,t);
	if( scale==LIS_SCALE_SYMM_DIAG && precon_type!=LIS_PRECON_TYPE_IS)
	{
		#pragma omp parallel for
		for(i=0;i<n;i++)
		{
			t->value[i] = t->value[i]/d->value[i];
		}
	}
	lis_vector_nrm2(t,&nrm2);
	solver->resid = nrm2;
	if( A->my_rank==0 )
	{
		if( err )
		{
			printf("lis_solve is %s(code=%d)\n",lis_returncode[err],err);
		}
		else
		{
			printf("lis_solve is normal end\n");
		}
	}
	if( precision==LIS_PRECISION_DOUBLE )
	{
		solver->iter2 = solver->iter;
	}

	lis_vector_destroy(t);
	lis_vector_destroy(d);
	lis_vector_destroy(xx);
	lis_precon_destroy(precon);
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_initial_residual"
int lis_solver_get_initial_residual(LIS_SOLVER solver, LIS_PRECON M, LIS_VECTOR t, LIS_VECTOR r, LIS_SCALAR *bnrm2)
{
	int			i;
	LIS_MATRIX	A;
	LIS_VECTOR	x,b,p,xx;
	LIS_SCALAR	nrm2;

	LIS_DEBUG_FUNC_IN;

	A  = solver->A;
	b  = solver->b;
	x  = solver->x;
	xx = solver->xx;

	/* ||b|| */
	lis_vector_nrm2(b,bnrm2);
	if( *bnrm2 == 0.0 )
	{
		*bnrm2 = 1.0;
	}
	else
	{
		*bnrm2 = 1.0 / *bnrm2;
	}
	/* Initial Residual */
	if( M==NULL )
	{
		p = r;
	}
	else
	{
		p = t;
	}

	if( !solver->options[LIS_OPTIONS_INITGUESS_ZEROS] )
	{
		#ifndef USE_QUAD_PRECISION
			lis_matvec(A,x,p);           /* p = Ax    */
			lis_vector_xpay(b,-1,p);     /* p = b - p */
			lis_vector_nrm2(p,&nrm2);
			nrm2 = nrm2 * *bnrm2;
			if( nrm2 <= solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN] )
			{
				solver->retcode = LIS_SUCCESS;
				solver->iter    = 1;
				solver->resid   = nrm2; 
				LIS_DEBUG_FUNC_OUT;
				return LIS_FAILS;
			}
		#else
			if( solver->precision==LIS_PRECISION_DOUBLE )
			{
				lis_matvec(A,x,p);           /* p = Ax    */
				lis_vector_xpay(b,-1,p);     /* p = b - p */
				lis_vector_nrm2(p,&nrm2);
				nrm2 = nrm2 * *bnrm2;
				if( nrm2 <= solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN] )
				{
					solver->retcode = LIS_SUCCESS;
					solver->iter    = 1;
					solver->resid   = nrm2; 
					LIS_DEBUG_FUNC_OUT;
					return LIS_FAILS;
				}
			}
			else
			{
				lis_matvec(A,xx,p);           /* p = Ax    */
				lis_vector_xpay(b,-1,p);     /* p = b - p */
				lis_vector_nrm2(p,&nrm2);
				nrm2 = nrm2 * *bnrm2;
				if( nrm2 <= solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN] )
				{
					solver->retcode = LIS_SUCCESS;
					solver->iter    = 1;
					solver->resid   = nrm2; 
					LIS_DEBUG_FUNC_OUT;
					return LIS_FAILS;
				}
				
				#pragma omp parallel for private(i)
				for(i=0;i<A->n;i++)
				{
					p->value_lo[i] = 0.0;
				}
				
			}
		#endif
	}
	else
	{
		#ifndef USE_QUAD_PRECISION
			lis_vector_copy(b,p);
		#else
			if( solver->precision==LIS_PRECISION_DOUBLE )
			{
				lis_vector_copy(b,p);
			}
			else
			{
				lis_vector_copyex_nm(b,p);
			}
		#endif
	}
	if( M!=NULL )
	{
		/* r = M^-1 * p */
		lis_psolve(solver, p, r);
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_optionC"
int lis_solver_set_optionC(LIS_SOLVER solver)
{
	LIS_ARGS	p;

	LIS_DEBUG_FUNC_IN;

	p = cmd_args->next;
	while( p!=cmd_args )
	{
		lis_solver_set_option2(p->arg1,p->arg2,solver);
		p = p->next;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option"
int lis_solver_set_option(char *text, LIS_SOLVER solver)
{
	LIS_ARGS	args,p;

	LIS_DEBUG_FUNC_IN;

	lis_text2args(text,&args);
	p = args->next;
	while( p!=args )
	{
		lis_solver_set_option2(p->arg1,p->arg2,solver);
		p = p->next;
	}
	lis_args_free(args);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option2"
int lis_solver_set_option2(char* arg1, char *arg2, LIS_SOLVER solver)
{
	int i;

	LIS_DEBUG_FUNC_IN;

	for(i=0;i<LIS_SOLVER_OPTION_LEN;i++)
	{
		if( strcmp(arg1, LIS_SOLVER_OPTNAME[i])==0 )
		{
			switch( LIS_SOLVER_OPTACT[i] )
			{
			case LIS_OPTIONS_FILE:
				break;
			case LIS_OPTIONS_HELP:
				break;
			case LIS_OPTIONS_VER:
				break;
			case LIS_OPTIONS_SOLVER:
				lis_solver_set_option_solver(arg2,solver);
				break;
			case LIS_OPTIONS_PRECON:
				lis_solver_set_option_precon(arg2,solver);
				break;
			case LIS_OPTIONS_SCALE:
				lis_solver_set_option_scale(arg2,solver);
				break;
			case LIS_OPTIONS_OUTPUT:
				lis_solver_set_option_print(arg2,solver);
				break;
			case LIS_OPTIONS_PSOLVER:
				lis_solver_set_option_psolver(arg2,solver);
				break;
			case LIS_OPTIONS_PPRECON:
				lis_solver_set_option_pprecon(arg2,solver);
				break;
			case LIS_OPTIONS_INITGUESS_ZEROS:
				lis_solver_set_option_truefalse(arg2,LIS_OPTIONS_INITGUESS_ZEROS,solver);
				break;
			case LIS_OPTIONS_ADDS:
				lis_solver_set_option_truefalse(arg2,LIS_OPTIONS_ADDS,solver);
				break;
			case LIS_OPTIONS_PRECISION:
				lis_solver_set_option_precision(arg2,LIS_OPTIONS_PRECISION,solver);
				break;
			case LIS_OPTIONS_USE_AT:
				lis_solver_set_option_truefalse(arg2,LIS_OPTIONS_USE_AT,solver);
				break;
			case LIS_OPTIONS_SAAMG_UNSYM:
				lis_solver_set_option_truefalse(arg2,LIS_OPTIONS_SAAMG_UNSYM,solver);
				break;
			default:
				if( LIS_SOLVER_OPTACT[i] < LIS_OPTIONS_LEN )
				{
					sscanf(arg2, "%d", &solver->options[LIS_SOLVER_OPTACT[i]]);
				}
				else
				{
					sscanf(arg2, "%lg", &solver->params[LIS_SOLVER_OPTACT[i]-LIS_OPTIONS_LEN]);
				}
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_solver"
int lis_solver_set_option_solver(char *argv, LIS_SOLVER solver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='9' )
	{
		sscanf(argv, "%d", &solver->options[LIS_OPTIONS_SOLVER]);
	}
	else
	{
		for(i=0;i<LIS_SOLVER_LEN;i++)
		{
			if( strcmp(argv,lis_solver_atoi[i])==0 )
			{
				solver->options[LIS_OPTIONS_SOLVER] = i+1;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_psolver"
int lis_solver_set_option_psolver(char *argv, LIS_SOLVER solver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='9' )
	{
		sscanf(argv, "%d", &solver->options[LIS_OPTIONS_PSOLVER]);
	}
	else
	{
		for(i=0;i<LIS_SOLVER_LEN;i++)
		{
			if( strcmp(argv,lis_solver_atoi[i])==0 )
			{
				solver->options[LIS_OPTIONS_PSOLVER] = i+1;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_precon"
int lis_solver_set_option_precon(char *argv, LIS_SOLVER solver)
{
	LIS_PRECON_REGISTER	preg;
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='9' )
	{
		sscanf(argv, "%d", &solver->options[LIS_OPTIONS_PRECON]);
	}
	else
	{
		for(i=0;i<LIS_PRECON_TYPE_LEN;i++)
		{
			if( strcmp(argv,lis_precon_atoi[i])==0 )
			{
				solver->options[LIS_OPTIONS_PRECON] = i;
				LIS_DEBUG_FUNC_OUT;
				return LIS_SUCCESS;
			}
		}
		for(i=0;i<precon_register_type-LIS_PRECON_TYPE_USERDEF;i++)
		{
			if( strcmp(argv,precon_register_top[i].name)==0 )
			{
				solver->options[LIS_OPTIONS_PRECON] = i+LIS_PRECON_TYPE_USERDEF;
				LIS_DEBUG_FUNC_OUT;
				return LIS_SUCCESS;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_pprecon"
int lis_solver_set_option_pprecon(char *argv, LIS_SOLVER solver)
{
	LIS_PRECON_REGISTER	preg;
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='9' )
	{
		sscanf(argv, "%d", &solver->options[LIS_OPTIONS_PPRECON]);
	}
	else
	{
		for(i=0;i<LIS_PRECON_TYPE_LEN;i++)
		{
			if( strcmp(argv,lis_precon_atoi[i])==0 )
			{
				solver->options[LIS_OPTIONS_PPRECON] = i;
				LIS_DEBUG_FUNC_OUT;
				return LIS_SUCCESS;
			}
		}
		for(i=0;i<precon_register_type-LIS_PRECON_TYPE_USERDEF;i++)
		{
			if( strcmp(argv,precon_register_top[i].name)==0 )
			{
				solver->options[LIS_OPTIONS_PPRECON] = i+LIS_PRECON_TYPE_USERDEF;
				LIS_DEBUG_FUNC_OUT;
				return LIS_SUCCESS;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_print"
int lis_solver_set_option_print(char *argv, LIS_SOLVER solver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='3' )
	{
		sscanf(argv, "%d", &solver->options[LIS_OPTIONS_OUTPUT]);
	}
	else
	{
		for(i=0;i<LIS_PRINT_LEN;i++)
		{
			if( strcmp(argv,lis_print_atoi[i])==0 )
			{
				solver->options[LIS_OPTIONS_OUTPUT] = i;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_scale"
int lis_solver_set_option_scale(char *argv, LIS_SOLVER solver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='2' )
	{
		sscanf(argv, "%d", &solver->options[LIS_OPTIONS_SCALE]);
	}
	else
	{
		for(i=0;i<LIS_SCALE_LEN;i++)
		{
			if( strcmp(argv,lis_scale_atoi[i])==0 )
			{
				solver->options[LIS_OPTIONS_SCALE] = i;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_truefalse"
int lis_solver_set_option_truefalse(char *argv, int opt, LIS_SOLVER solver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='1' )
	{
		sscanf(argv, "%d", &solver->options[opt]);
	}
	else
	{
		for(i=0;i<LIS_TRUEFALSE_LEN;i++)
		{
			if( strcmp(argv,lis_truefalse_atoi[i])==0 )
			{
				solver->options[opt] = i;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_set_option_precision"
int lis_solver_set_option_precision(char *argv, int opt, LIS_SOLVER solver)
{
	int  i;

	LIS_DEBUG_FUNC_IN;

	if( argv[0]>='0' && argv[0]<='1' )
	{
		sscanf(argv, "%d", &solver->options[opt]);
	}
	else
	{
		for(i=0;i<LIS_PRECISION_LEN;i++)
		{
			if( strcmp(argv,lis_precision_atoi[i])==0 )
			{
				solver->options[opt] = i;
				break;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_iters"
int lis_solver_get_iters(LIS_SOLVER solver, int *iters)
{
	LIS_DEBUG_FUNC_IN;

	*iters = solver->iter;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_times"
int lis_solver_get_times(LIS_SOLVER solver, double *times, double *itimes, double *ptimes)
{
	LIS_DEBUG_FUNC_IN;

	*times  = solver->times;
	if( itimes ) *itimes = solver->itimes;
	if( ptimes ) *ptimes = solver->ptimes;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_residualnorm"
int lis_solver_get_residualnorm(LIS_SOLVER solver, LIS_REAL *residual)
{
	LIS_DEBUG_FUNC_IN;

	*residual  = solver->resid;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_solver"
int lis_solver_get_solver(LIS_SOLVER solver, int *nsol)
{
	LIS_DEBUG_FUNC_IN;

	*nsol = solver->options[LIS_OPTIONS_SOLVER];

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_solver_get_status"
int lis_solver_get_status(LIS_SOLVER solver, int *status)
{
	LIS_DEBUG_FUNC_IN;

	*status = solver->retcode;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

int lis_get_residual_history(LIS_VECTOR v)
{
	int		i,n;

	n = _min(v->n,lis_residual_history_count);
	for(i=0;i<n;i++)
	{
		v->value[i] = lis_residual_history->value[i];
	}
	return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_get_solvername"
int lis_get_solvername(int solver, char *name)
{
	LIS_DEBUG_FUNC_IN;

	if( solver < 1 || solver > LIS_SOLVERS_LEN )
	{
		name = NULL;
		return LIS_FAILS;
	}
	strcpy(name,lis_solvername[solver]);

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

