/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include <stdarg.h>
#ifdef USE_SSE2
	#include <emmintrin.h>
#endif
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

#define NWORK				4
/************************************************
 * lis_cg_check_params
 * lis_cg_malloc_work
 * lis_cg
 ************************************************/
#undef __FUNC__
#define __FUNC__ "lis_cg_check_params"
int lis_cg_check_params(LIS_SOLVER solver)
{
	LIS_DEBUG_FUNC_IN;
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_cg_malloc_work"
int lis_cg_malloc_work(LIS_SOLVER solver)
{
	LIS_VECTOR	*work;
	int			i,j,worklen,err;

	LIS_DEBUG_FUNC_IN;

	worklen = NWORK;
	work    = (LIS_VECTOR *)lis_malloc( worklen*sizeof(LIS_VECTOR) );
	if( work==NULL )
	{
		LIS_SETERR_MEM(worklen*sizeof(LIS_VECTOR));
		return LIS_ERR_OUT_OF_MEMORY;
	}
	if( solver->precision==LIS_PRECISION_DEFAULT )
	{
		for(i=0;i<worklen;i++)
		{
			err = lis_vector_duplicate(solver->A,&work[i]);
			if( err ) break;
		}
	}
	else
	{
		for(i=0;i<worklen;i++)
		{
			err = lis_vector_duplicateex(LIS_PRECISION_QUAD,solver->A,&work[i]);
			if( err ) break;
		}
	}
	if( i<worklen )
	{
		for(j=0;j<i;j++) lis_vector_destroy(work[j]);
		lis_free(work);
		return err;
	}
	solver->worklen = worklen;
	solver->work    = work;

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_cg"
int lis_cg(LIS_SOLVER solver)
{
	LIS_MATRIX A;
	LIS_PRECON M;
	LIS_VECTOR b,x;
	LIS_VECTOR r,z,p,q;
	LIS_SCALAR alpha, beta, rho, rho_old;
	LIS_SCALAR bnrm2, nrm2, tol, dot_pq;
	int iter,maxiter,n,output;
	double ptimes, times;

	LIS_DEBUG_FUNC_IN;

	A       = solver->A;
	M       = solver->precon;
	b       = solver->b;
	x       = solver->x;
	n       = A->n;
	maxiter = solver->options[LIS_OPTIONS_MAXITER];
	output  = solver->options[LIS_OPTIONS_OUTPUT];
	tol     = solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN];
	ptimes  = 0.0;


	z       = solver->work[0];
	q       = solver->work[1];
	r       = solver->work[2];
	p       = solver->work[3];
	rho_old = (LIS_SCALAR)1.0;
	beta    = (LIS_SCALAR)0.0;


	if( lis_solver_get_initial_residual(solver,NULL,NULL,r,&bnrm2) )
	{
		LIS_DEBUG_FUNC_OUT;
		return LIS_SUCCESS;
	}
	lis_vector_set_all(0.0,p);
	
	for( iter=1; iter<=maxiter; iter++ )
	{
		/* z = M^-1 * r */
		times = lis_wtime();
		lis_psolve(solver,r,z);
		ptimes += lis_wtime() - times;

		/* rho = <r,z> */
		lis_vector_dot(r,z,&rho);

		/* beta = rho / rho_old */
		beta = rho / rho_old;

		/* p = z + beta*p       */
		lis_vector_xpay(z,beta,p);
		
		/* q = Ap */
		lis_matvec(A,p,q);
		
		/* dot_pq = <p,q> */
		lis_vector_dot(p,q,&dot_pq);

		/* breakdown check */
		if( dot_pq==0.0 )
		{
			solver->retcode   = LIS_BREAKDOWN;
			solver->iter      = iter;
			solver->resid     = nrm2;
			LIS_DEBUG_FUNC_OUT;
			return LIS_BREAKDOWN;
		}
		
		/* alpha = rho / dot_pq */
		alpha = rho / dot_pq;
		
		/* x = x + alpha*p */
		lis_vector_axpy(alpha,p,x);
		
		/* r = r - alpha*q */
		lis_vector_axpy(-alpha,q,r);

		/* convergence check */
		lis_vector_nrm2(r,&nrm2);
		nrm2 = nrm2 * bnrm2;

		if( output )
		{
			if( output & LIS_PRINT_MEM ) solver->residual[iter-1] = nrm2;
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2);
		}
		
		if( tol >= nrm2 )
		{
			solver->retcode    = LIS_SUCCESS;
			solver->iter       = iter;
			solver->resid      = nrm2;
			solver->ptimes     = ptimes;
			LIS_DEBUG_FUNC_OUT;
			return LIS_SUCCESS;
		}
		rho_old = rho;
	}

	solver->retcode   = LIS_MAXITER;
	solver->iter      = iter;
	solver->resid     = nrm2;
	LIS_DEBUG_FUNC_OUT;
	return LIS_MAXITER;
}

#ifdef USE_QUAD_PRECISION
#undef __FUNC__
#define __FUNC__ "lis_cg_quad"
int lis_cg_quad(LIS_SOLVER solver)
{
	LIS_MATRIX A;
	LIS_PRECON M;
	LIS_VECTOR b,x;
	LIS_VECTOR r,z,p,q;
	LIS_SCALAR_M alpha, beta, rho, rho_old;
	LIS_SCALAR_M dot_pq, nrm2;
	LIS_SCALAR	bnrm2, tol;
	int iter,maxiter,n,output;
	double ptimes, times;

	LIS_DEBUG_FUNC_IN;

	A       = solver->A;
	M       = solver->precon;
	b       = solver->b;
	x       = solver->x;
	n       = A->n;
	maxiter = solver->options[LIS_OPTIONS_MAXITER];
	output  = solver->options[LIS_OPTIONS_OUTPUT];
	tol     = solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN];
	ptimes  = 0.0;

	LIS_QUAD_SCALAR_MALLOC(alpha,0,1);
	LIS_QUAD_SCALAR_MALLOC(beta,1,1);
	LIS_QUAD_SCALAR_MALLOC(rho,2,1);
	LIS_QUAD_SCALAR_MALLOC(rho_old,3,1);
	LIS_QUAD_SCALAR_MALLOC(dot_pq,4,1);
	LIS_QUAD_SCALAR_MALLOC(nrm2,5,1);
	
	z       = solver->work[0];
	q       = solver->work[1];
	r       = solver->work[2];
	p       = solver->work[3];
	rho_old.hi[0] = 1.0;
	rho_old.lo[0] = 0.0;
	beta.hi[0] = 0.0;
	beta.lo[0] = 0.0;

	if( lis_solver_get_initial_residual(solver,NULL,NULL,r,&bnrm2) )
	{
		LIS_DEBUG_FUNC_OUT;
		return LIS_SUCCESS;
	}
	lis_vector_set_allex_nm(0.0,p);
	
	for( iter=1; iter<=maxiter; iter++ )
	{
		/* z = M^-1 * r */
		times = lis_wtime();
		lis_psolve(solver,r,z);
		ptimes += lis_wtime() - times;

		/* rho = <r,z> */
		lis_vector_dotex_mmm(r,z,&rho);

		/* beta = rho / rho_old */
		lisqp_div((LIS_QDOUBLE *)beta.hi,(LIS_QDOUBLE *)rho.hi,(LIS_QDOUBLE *)rho_old.hi);

		/* p = z + beta*p       */
		lis_vector_xpayex_mmm(z,beta,p);

		/* q = Ap */
		lis_matvec(A,p,q);
		
		/* dot_pq = <p,q> */
		lis_vector_dotex_mmm(p,q,&dot_pq);

		/* breakdown check */
		if( dot_pq.hi[0]==0.0 && dot_pq.lo[0]==0.0 )
		{
			solver->retcode   = LIS_BREAKDOWN;
			solver->iter      = iter;
			solver->resid     = nrm2.hi[0];
			LIS_DEBUG_FUNC_OUT;
			return LIS_BREAKDOWN;
		}
		
		/* alpha = rho / dot_pq */
		lisqp_div((LIS_QDOUBLE *)alpha.hi,(LIS_QDOUBLE *)rho.hi,(LIS_QDOUBLE *)dot_pq.hi);
		
		/* x = x + alpha*p */
		lis_vector_axpyex_mmm(alpha,p,x);
		
		/* r = r - alpha*q */
		lisqp_minus((LIS_QDOUBLE *)alpha.hi);
		lis_vector_axpyex_mmm(alpha,q,r);

		/* convergence check */

		lis_vector_nrm2ex_mm(r,&nrm2);
		lisqp_mul_dd_d((LIS_QDOUBLE *)nrm2.hi,(LIS_QDOUBLE *)nrm2.hi,bnrm2);
		if( output )
		{
			if( output & LIS_PRINT_MEM ) solver->residual[iter-1] = nrm2.hi[0];
			if( output & LIS_PRINT_OUT && A->my_rank==0 ) printf("iter: %5d  residual = %e\n", iter, nrm2.hi[0]);
		}
		
		if( (tol > nrm2.hi[0]) || (tol==nrm2.hi[0] && nrm2.lo[0]<=0.0) )
		{
			solver->retcode    = LIS_SUCCESS;
			solver->iter       = iter;
			solver->resid      = nrm2.hi[0];
			solver->ptimes     = ptimes;
			LIS_DEBUG_FUNC_OUT;
			return LIS_SUCCESS;
		}
		rho_old.hi[0] = rho.hi[0];
		rho_old.lo[0] = rho.lo[0];
	}

	solver->retcode   = LIS_MAXITER;
	solver->iter      = iter;
	solver->resid     = nrm2.hi[0];
	LIS_DEBUG_FUNC_OUT;
	return LIS_MAXITER;
}

#undef __FUNC__
#define __FUNC__ "lis_cg_switch"
int lis_cg_switch(LIS_SOLVER solver)
{
	LIS_MATRIX A;
	LIS_PRECON M;
	LIS_VECTOR b,x;
	LIS_VECTOR r,z,p,q;
	LIS_SCALAR_M alpha, beta, rho, rho_old;
	LIS_SCALAR_M dot_pq, nrm2;
	LIS_SCALAR	bnrm2, tol,tol2;
	int iter,iter2,maxiter,maxiter2,n,output;
	double ptimes, times;

	LIS_DEBUG_FUNC_IN;

	A       = solver->A;
	M       = solver->precon;
	b       = solver->b;
	x       = solver->x;
	n       = A->n;
	maxiter  = solver->options[LIS_OPTIONS_MAXITER];
	maxiter2 = solver->options[LIS_OPTIONS_SWITCH_MAXITER];
	output   = solver->options[LIS_OPTIONS_OUTPUT];
	tol      = solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN];
	tol2     = solver->params[LIS_PARAMS_SWITCH_RESID-LIS_OPTIONS_LEN];
	ptimes  = 0.0;

	LIS_QUAD_SCALAR_MALLOC(alpha,0,1);
	LIS_QUAD_SCALAR_MALLOC(beta,1,1);
	LIS_QUAD_SCALAR_MALLOC(rho,2,1);
	LIS_QUAD_SCALAR_MALLOC(rho_old,3,1);
	LIS_QUAD_SCALAR_MALLOC(dot_pq,4,1);
	LIS_QUAD_SCALAR_MALLOC(nrm2,5,1);
	
	z       = solver->work[0];
	q       = solver->work[1];
	r       = solver->work[2];
	p       = solver->work[3];
	rho_old.hi[0] = 1.0;
	rho_old.lo[0] = 0.0;
	beta.hi[0] = 0.0;
	beta.lo[0] = 0.0;

	if( lis_solver_get_initial_residual(solver,NULL,NULL,r,&bnrm2) )
	{
		LIS_DEBUG_FUNC_OUT;
		return LIS_SUCCESS;
	}
	lis_vector_set_allex_nm(0.0,p);
	r->precision = LIS_PRECISION_DEFAULT;
	p->precision = LIS_PRECISION_DEFAULT;
	
	for( iter=1; iter<=maxiter2; iter++ )
	{
		/* z = M^-1 * r */
		times = lis_wtime();
		lis_psolve(solver,r,z);
		ptimes += lis_wtime() - times;

		/* rho = <r,z> */
		lis_vector_dot(r,z,&rho.hi[0]);

		/* beta = rho / rho_old */
		beta.hi[0] = rho.hi[0] / rho_old.hi[0];

		/* p = z + beta*p       */
		lis_vector_xpay(z,beta.hi[0],p);
		
		/* q = Ap */
		lis_matvec(A,p,q);
		
		/* dot_pq = <p,q> */
		lis_vector_dot(p,q,&dot_pq.hi[0]);

		/* breakdown check */
		if( dot_pq.hi[0]==0.0 )
		{
			solver->retcode   = LIS_BREAKDOWN;
			solver->iter      = iter;
			solver->iter2     = iter;
			solver->resid     = nrm2.hi[0];
			LIS_DEBUG_FUNC_OUT;
			return LIS_BREAKDOWN;
		}
		
		/* alpha = rho / dot_pq */
		alpha.hi[0] = rho.hi[0] / dot_pq.hi[0];
		
		/* x = x + alpha*p */
		lis_vector_axpy(alpha.hi[0],p,x);
		
		/* r = r - alpha*q */
		lis_vector_axpy(-alpha.hi[0],q,r);

		/* convergence check */
		lis_vector_nrm2(r,&nrm2.hi[0]);
		nrm2.hi[0] = nrm2.hi[0] * bnrm2;

		if( output )
		{
			if( output & LIS_PRINT_MEM ) solver->residual[iter-1] = nrm2.hi[0];
			if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2.hi[0]);
		}
		
		if( tol2 >= nrm2.hi[0] )
		{
			solver->iter       = iter;
			solver->iter2     = iter;
			solver->ptimes     = ptimes;
			break;
		}
		rho_old.hi[0] = rho.hi[0];
	}

	r->precision = LIS_PRECISION_QUAD;
	p->precision = LIS_PRECISION_QUAD;
	
	solver->options[LIS_OPTIONS_INITGUESS_ZEROS] = LIS_FALSE;
	lis_vector_copyex_mn(x,solver->xx);
	rho_old.hi[0] = 1.0;
	beta.hi[0] = 0.0;
	if( lis_solver_get_initial_residual(solver,NULL,NULL,r,&bnrm2) )
	{
		LIS_DEBUG_FUNC_OUT;
		return LIS_SUCCESS;
	}
	lis_vector_set_allex_nm(0.0,p);
	
	for( iter2=iter+1; iter2<=maxiter; iter2++ )
	{
		/* z = M^-1 * r */
		times = lis_wtime();
		lis_psolve(solver,r,z);
		ptimes += lis_wtime() - times;

		/* rho = <r,z> */
		lis_vector_dotex_mmm(r,z,&rho);

		/* beta = rho / rho_old */
		lisqp_div((LIS_QDOUBLE *)beta.hi,(LIS_QDOUBLE *)rho.hi,(LIS_QDOUBLE *)rho_old.hi);

		/* p = z + beta*p       */
		lis_vector_xpayex_mmm(z,beta,p);

		/* q = Ap */
		lis_matvec(A,p,q);
		
		/* dot_pq = <p,q> */
		lis_vector_dotex_mmm(p,q,&dot_pq);

		/* breakdown check */
		if( dot_pq.hi[0]==0.0 && dot_pq.lo[0]==0.0 )
		{
			solver->retcode   = LIS_BREAKDOWN;
			solver->iter      = iter2;
			solver->resid     = nrm2.hi[0];
			LIS_DEBUG_FUNC_OUT;
			return LIS_BREAKDOWN;
		}
		
		/* alpha = rho / dot_pq */
		lisqp_div((LIS_QDOUBLE *)alpha.hi,(LIS_QDOUBLE *)rho.hi,(LIS_QDOUBLE *)dot_pq.hi);
		
		/* x = x + alpha*p */
		lis_vector_axpyex_mmm(alpha,p,x);
		
		/* r = r - alpha*q */
		lisqp_minus((LIS_QDOUBLE *)alpha.hi);
		lis_vector_axpyex_mmm(alpha,q,r);

		/* convergence check */

		lis_vector_nrm2ex_mm(r,&nrm2);
		lisqp_mul_dd_d((LIS_QDOUBLE *)nrm2.hi,(LIS_QDOUBLE *)nrm2.hi,bnrm2);
		if( output )
		{
			if( output & LIS_PRINT_MEM ) solver->residual[iter2-1] = nrm2.hi[0];
			if( output & LIS_PRINT_OUT && A->my_rank==0 ) printf("iter: %5d  residual = %e\n", iter2, nrm2.hi[0]);
		}
		
		if( (tol > nrm2.hi[0]) || (tol==nrm2.hi[0] && nrm2.lo[0]<=0.0) )
		{
			solver->retcode    = LIS_SUCCESS;
			solver->iter       = iter2;
			solver->resid      = nrm2.hi[0];
			solver->ptimes     = ptimes;
			LIS_DEBUG_FUNC_OUT;
			return LIS_SUCCESS;
		}
		rho_old.hi[0] = rho.hi[0];
		rho_old.lo[0] = rho.lo[0];
	}

	solver->retcode   = LIS_MAXITER;
	solver->iter      = iter2;
	solver->resid     = nrm2.hi[0];
	LIS_DEBUG_FUNC_OUT;
	return LIS_MAXITER;
}

#undef __FUNC__
#define __FUNC__ "lis_cg_periodic"
int lis_cg_periodic(LIS_SOLVER solver)
{
	LIS_MATRIX A;
	LIS_PRECON M;
	LIS_VECTOR b,x;
	LIS_VECTOR r,z,p,q;
	LIS_SCALAR_M alpha, beta, rho, rho_old;
	LIS_SCALAR_M dot_pq, nrm2;
	LIS_SCALAR	bnrm2, tol;
	int iter,iter2,maxiter,n,output,qnum,qinter;
	double ptimes, times;

	LIS_DEBUG_FUNC_IN;

	A       = solver->A;
	M       = solver->precon;
	b       = solver->b;
	x       = solver->x;
	n       = A->n;
	maxiter  = solver->options[LIS_OPTIONS_MAXITER];
	output   = solver->options[LIS_OPTIONS_OUTPUT];
	qnum     = solver->options[LIS_OPTIONS_PERIODIC_NUM];
	qinter   = solver->options[LIS_OPTIONS_PERIODIC_INTERVAL];
	tol      = solver->params[LIS_PARAMS_RESID-LIS_OPTIONS_LEN];
	ptimes   = 0.0;
	iter2    = 0;

	LIS_QUAD_SCALAR_MALLOC(alpha,0,1);
	LIS_QUAD_SCALAR_MALLOC(beta,1,1);
	LIS_QUAD_SCALAR_MALLOC(rho,2,1);
	LIS_QUAD_SCALAR_MALLOC(rho_old,3,1);
	LIS_QUAD_SCALAR_MALLOC(dot_pq,4,1);
	LIS_QUAD_SCALAR_MALLOC(nrm2,5,1);
	
	z       = solver->work[0];
	q       = solver->work[1];
	r       = solver->work[2];
	p       = solver->work[3];
	rho_old.hi[0] = 1.0;
	rho_old.lo[0] = 0.0;
	beta.hi[0] = 0.0;
	beta.lo[0] = 0.0;

	if( lis_solver_get_initial_residual(solver,NULL,NULL,r,&bnrm2) )
	{
		LIS_DEBUG_FUNC_OUT;
		return LIS_SUCCESS;
	}
	lis_vector_set_allex_nm(0.0,p);
	r->precision = LIS_PRECISION_DEFAULT;
	p->precision = LIS_PRECISION_DEFAULT;
	
	for( iter=1; iter<=maxiter; iter++ )
	{
		if( (iter-1)%qinter==0 )
		{
			memset(r->value_lo,0,r->np*sizeof(LIS_SCALAR));
			memset(p->value_lo,0,r->np*sizeof(LIS_SCALAR));
			memset(x->value_lo,0,r->np*sizeof(LIS_SCALAR));
			rho_old.lo[0] = 0.0;
		}
		if( (iter-1)%qinter<qnum )
		{
			r->precision = LIS_PRECISION_QUAD;
			p->precision = LIS_PRECISION_QUAD;
		
			/* z = M^-1 * r */
			times = lis_wtime();
			lis_psolve(solver,r,z);
			ptimes += lis_wtime() - times;

			/* rho = <r,z> */
			lis_vector_dotex_mmm(r,z,&rho);

			/* beta = rho / rho_old */
			lisqp_div((LIS_QDOUBLE *)beta.hi,(LIS_QDOUBLE *)rho.hi,(LIS_QDOUBLE *)rho_old.hi);

			/* p = z + beta*p       */
			lis_vector_xpayex_mmm(z,beta,p);

			/* q = Ap */
			lis_matvec(A,p,q);
			
			/* dot_pq = <p,q> */
			lis_vector_dotex_mmm(p,q,&dot_pq);

			/* breakdown check */
			if( dot_pq.hi[0]==0.0 && dot_pq.lo[0]==0.0 )
			{
				solver->retcode   = LIS_BREAKDOWN;
				solver->iter      = iter;
				solver->iter2     = iter2;
				solver->resid     = nrm2.hi[0];
				LIS_DEBUG_FUNC_OUT;
				return LIS_BREAKDOWN;
			}
			
			/* alpha = rho / dot_pq */
			lisqp_div((LIS_QDOUBLE *)alpha.hi,(LIS_QDOUBLE *)rho.hi,(LIS_QDOUBLE *)dot_pq.hi);
			
			/* x = x + alpha*p */
			lis_vector_axpyex_mmm(alpha,p,x);
			
			/* r = r - alpha*q */
			lisqp_minus((LIS_QDOUBLE *)alpha.hi);
			lis_vector_axpyex_mmm(alpha,q,r);

			/* convergence check */

			lis_vector_nrm2ex_mm(r,&nrm2);
			lisqp_mul_dd_d((LIS_QDOUBLE *)nrm2.hi,(LIS_QDOUBLE *)nrm2.hi,bnrm2);
			if( output )
			{
				if( output & LIS_PRINT_MEM ) solver->residual[iter-1] = nrm2.hi[0];
				if( output & LIS_PRINT_OUT && A->my_rank==0 ) printf("iter: %5d  residual = %e\n", iter, nrm2.hi[0]);
			}
			
			if( (tol > nrm2.hi[0]) || (tol==nrm2.hi[0] && nrm2.lo[0]<=0.0) )
			{
				solver->retcode   = LIS_SUCCESS;
				solver->iter      = iter;
				solver->iter2     = iter2;
				solver->resid     = nrm2.hi[0];
				solver->ptimes    = ptimes;
				LIS_DEBUG_FUNC_OUT;
				return LIS_SUCCESS;
			}
			rho_old.hi[0] = rho.hi[0];
			rho_old.lo[0] = rho.lo[0];
		}
		else
		{
			iter2++;
			r->precision = LIS_PRECISION_DEFAULT;
			p->precision = LIS_PRECISION_DEFAULT;

			/* z = M^-1 * r */
			times = lis_wtime();
			lis_psolve(solver,r,z);
			ptimes += lis_wtime() - times;

			/* rho = <r,z> */
			lis_vector_dot(r,z,&rho.hi[0]);

			/* beta = rho / rho_old */
			beta.hi[0] = rho.hi[0] / rho_old.hi[0];

			/* p = z + beta*p       */
			lis_vector_xpay(z,beta.hi[0],p);
			
			/* q = Ap */
			lis_matvec(A,p,q);
			
			/* dot_pq = <p,q> */
			lis_vector_dot(p,q,&dot_pq.hi[0]);

			/* breakdown check */
			if( dot_pq.hi[0]==0.0 )
			{
				solver->retcode   = LIS_BREAKDOWN;
				solver->iter      = iter;
				solver->iter2     = iter2;
				solver->resid     = nrm2.hi[0];
				LIS_DEBUG_FUNC_OUT;
				return LIS_BREAKDOWN;
			}
			
			/* alpha = rho / dot_pq */
			alpha.hi[0] = rho.hi[0] / dot_pq.hi[0];
			
			/* x = x + alpha*p */
			lis_vector_axpy(alpha.hi[0],p,x);
			
			/* r = r - alpha*q */
			lis_vector_axpy(-alpha.hi[0],q,r);

			/* convergence check */
			lis_vector_nrm2(r,&nrm2.hi[0]);
			nrm2.hi[0] = nrm2.hi[0] * bnrm2;

			if( output )
			{
				if( output & LIS_PRINT_MEM ) solver->residual[iter-1] = nrm2.hi[0];
				if( output & LIS_PRINT_OUT ) printf("iter: %5d  residual = %e\n", iter, nrm2.hi[0]);
			}
			
			if( tol >= nrm2.hi[0] )
			{
				solver->retcode    = LIS_SUCCESS;
				solver->iter       = iter;
				solver->iter2      = iter2;
				solver->resid      = nrm2.hi[0];
				solver->ptimes     = ptimes;
				LIS_DEBUG_FUNC_OUT;
				return LIS_SUCCESS;
			}
			rho_old.hi[0] = rho.hi[0];
		}
	}
	solver->retcode   = LIS_MAXITER;
	solver->iter      = iter;
	solver->iter2     = iter2;
	solver->resid     = nrm2.hi[0];
	LIS_DEBUG_FUNC_OUT;
	return LIS_MAXITER;
}
#endif
