/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_vector_dot			v   <- x' * y
 * lis_vector_nrm2			v   <- ||x||_2
 * lis_vector_sum			v   <- sum x_i
 ************************************************/

/**********************/
/* v <- x' * y        */
/**********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_dot"
int lis_vector_dot(LIS_VECTOR vx, LIS_VECTOR vy, LIS_SCALAR *val)
{
	int i,n,nprocs,my_rank;
	LIS_SCALAR tmp;
	LIS_SCALAR dot;
	LIS_SCALAR *x,*y;
	#ifdef USE_MPI
		MPI_Comm   comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	n = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x      = vx->value;
	y      = vy->value;
	#ifdef USE_MPI
		comm   = vx->comm;
	#endif
	#ifdef _OPENMP
		nprocs = omp_get_max_threads();
		#pragma omp parallel private(i,tmp,my_rank)
		{
			my_rank = omp_get_thread_num();
			tmp     = 0.0;
		    #pragma cdir nodep
			#pragma omp for
			for(i=0; i<n; i++)
			{
				tmp += x[i]*y[i];
			}
			lis_vec_tmp[my_rank*LIS_VEC_TMP_PADD] = tmp;
		}
		dot = 0.0;
		for(i=0;i<nprocs;i++)
		{
			dot += lis_vec_tmp[i*LIS_VEC_TMP_PADD];
		}
	#else
		dot  = 0.0;
	    #pragma cdir nodep
		for(i=0; i<n; i++)
		{
			dot += x[i]*y[i];
		}
	#endif
	#ifdef USE_MPI
		MPI_Allreduce(&dot,&tmp,1,MPI_DOUBLE,MPI_SUM,comm);
		*val = tmp;
	#else
		*val = dot;
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

/**********************/
/* v <- ||x||_2       */
/**********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_nrm2"
int lis_vector_nrm2(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n,nprocs,my_rank;
	LIS_SCALAR tmp;
	LIS_SCALAR dot;
	LIS_SCALAR *x;
	#ifdef USE_MPI
		MPI_Comm   comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	n      = vx->n;

	x      = vx->value;
	#ifdef USE_MPI
		comm   = vx->comm;
	#endif
	#ifdef _OPENMP
		nprocs = omp_get_max_threads();
		#pragma omp parallel private(i,tmp,my_rank)
		{
			my_rank = omp_get_thread_num();
			tmp     = 0.0;
		    #pragma cdir nodep
			#pragma omp for
			for(i=0; i<n; i++)
			{
				tmp += x[i]*x[i];
			}
			lis_vec_tmp[my_rank*LIS_VEC_TMP_PADD] = tmp;
		}
		dot = 0.0;
		for(i=0;i<nprocs;i++)
		{
			dot += lis_vec_tmp[i*LIS_VEC_TMP_PADD];
		}
	#else
		dot  = 0.0;
	    #pragma cdir nodep
		for(i=0; i<n; i++)
		{
			dot += x[i]*x[i];
		}
	#endif
	#ifdef USE_MPI
		MPI_Allreduce(&dot,&tmp,1,MPI_DOUBLE,MPI_SUM,comm);
		*val = sqrt(tmp);
	#else
		*val = sqrt(dot);
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

/**********************/
/* v <- sum x_i       */
/**********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_sum"
int lis_vector_sum(LIS_VECTOR vx, LIS_SCALAR *val)
{
	int i,n,nprocs,my_rank;
	LIS_SCALAR tmp;
	LIS_SCALAR sum;
	LIS_SCALAR *x;
	#ifdef USE_MPI
		MPI_Comm   comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	n   = vx->n;

	x      = vx->value;
	#ifdef USE_MPI
		comm   = vx->comm;
	#endif
	sum    = 0.0;
	#ifdef _OPENMP
		nprocs = omp_get_max_threads();
		#pragma omp parallel private(i,tmp,my_rank)
		{
			my_rank = omp_get_thread_num();
			tmp     = 0.0;
		    #pragma cdir nodep
			#pragma omp for
			for(i=0; i<n; i++)
			{
				tmp += x[i];
			}
			lis_vec_tmp[my_rank*LIS_VEC_TMP_PADD] = tmp;
		}
		for(i=0;i<nprocs;i++)
		{
			sum += lis_vec_tmp[i*LIS_VEC_TMP_PADD];
		}
	#else
	    #pragma cdir nodep
		for(i=0; i<n; i++)
		{
			sum += x[i];
		}
	#endif
	#ifdef USE_MPI
		MPI_Allreduce(&sum,val,1,MPI_DOUBLE,MPI_SUM,comm);
	#else
		*val = sum;
	#endif

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#ifdef USE_QUAD_PRECISION
#undef __FUNC__
#define __FUNC__ "lis_vector_dotex"
int lis_vector_dotex(LIS_VECTOR vx, LIS_VECTOR vy, LIS_MSCALAR *val)
{
	int i,n,is,ie,nprocs,my_rank;
	LIS_SCALAR dot,tmp;
	LIS_SCALAR *x,*y;
	LIS_MSCALAR dotm,tmpm,xy;
	LIS_MSCALAR *xm,*ym;
	#ifdef USE_MPI
		MPI_Comm   comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	n      = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	if( vx->precision>=LIS_PRECISION_QUAD && vy->precision>=LIS_PRECISION_QUAD )
	{
		xm = (LIS_MSCALAR *)vx->value;
		ym = (LIS_MSCALAR *)vy->value;
		#ifdef USE_MPI
			comm   = vx->comm;
		#endif
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			#pragma omp parallel private(i,tmpm,xy,is,ie,my_rank)
			{
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
				lisqp_zero(&tmpm);
				for(i=is;i<ie;i++)
				{
					lisqp_mul(&xy,&xm[i],&ym[i]);
					lisqp_add(&tmpm,&tmpm,&xy);
				}
				((LIS_MSCALAR *)lis_vec_tmp)[my_rank*LIS_VEC_TMP_PADD] = tmpm;
			}
			lisqp_zero(&dotm);
			for(i=0;i<nprocs;i++)
			{
				lisqp_add(&dotm,&dotm,&((LIS_MSCALAR *)lis_vec_tmp)[i*LIS_VEC_TMP_PADD]);
			}
		#else
			lisqp_zero(&dotm);
			for(i=0;i<n;i++)
			{
				lisqp_mul(&xy,&xm[i],&ym[i]);
				lisqp_add(&dotm,&dotm,&xy);
			}
		#endif
		#ifdef _USE_MPI
			MPI_Allreduce(&dotm,&tmpm,1,LIS_MPI_MSCALAR,LIS_MPI_MSUM,comm);
			*val = tmpm;
		#else
			*val = dotm;
		#endif
	}
	else if( vx->precision==LIS_PRECISION_DEFAULT && vy->precision==LIS_PRECISION_DEFAULT )
	{
		x   = vx->value;
		y   = vy->value;
		#ifdef USE_MPI
			comm   = vx->comm;
		#endif
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			#pragma omp parallel private(i,tmp,my_rank)
			{
				my_rank = omp_get_thread_num();
				tmp     = (LIS_SCALAR)0.0;
				#pragma omp for
				for(i=0;i<n;i++)
				{
					tmp += x[i]*y[i];
				}
				lis_vec_tmp[my_rank*LIS_VEC_TMP_PADD] = tmp;
			}
			dot = 0.0;
			for(i=0;i<nprocs;i++)
			{
				dot += lis_vec_tmp[i*LIS_VEC_TMP_PADD];
			}
		#else
			dot = 0.0;
			for(i=0; i<n; i++)
			{
				dot += x[i]*y[i];
			}
		#endif
		#ifdef USE_MPI
			MPI_Allreduce(&dot,&tmp,1,MPI_DOUBLE,MPI_SUM,comm);
			(*val).hi = tmp;
		#else
			(*val).hi = dot;
		#endif
	}
	else
	{
		if( vx->precision!=LIS_PRECISION_DOUBLE )
		{
			xm = (LIS_MSCALAR *)vx->value;
			y  = vy->value;
			#ifdef USE_MPI
				comm   = vx->comm;
			#endif
			#ifdef _OPENMP
				nprocs = omp_get_max_threads();
				#pragma omp parallel private(i,tmpm,xy,my_rank)
				{
					my_rank = omp_get_thread_num();
					LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
					lisqp_zero(&tmpm);
					for(i=is;i<ie;i++)
					{
						lisqp_mul_dd_d(&xy,&xm[i],y[i]);
						lisqp_add(&tmpm,&tmpm,&xy);
					}
					((LIS_MSCALAR *)lis_vec_tmp)[my_rank*LIS_VEC_TMP_PADD] = tmpm;
				}
				lisqp_zero(&dotm);
				for(i=0;i<nprocs;i++)
				{
					lisqp_add(&dotm,&dotm,&((LIS_MSCALAR *)lis_vec_tmp)[i*LIS_VEC_TMP_PADD]);
				}
			#else
				lisqp_zero(&dotm);
				for(i=0;i<n;i++)
				{
					lisqp_mul_dd_d(&xy,&xm[i],y[i]);
					lisqp_add(&dotm,&dotm,&xy);
				}
			#endif
			#ifdef USE_MPI
				MPI_Allreduce(&dotm,&tmpm,1,LIS_MPI_MSCALAR,LIS_MPI_MSUM,comm);
				*val = tmpm;
			#else
				*val = dotm;
			#endif
		}
		else
		{
			x  = vx->value;
			ym = (LIS_MSCALAR *)vy->value;
			#ifdef USE_MPI
				comm   = vx->comm;
			#endif
			#ifdef _OPENMP
				nprocs = omp_get_max_threads();
				#pragma omp parallel private(i,tmpm,xy,my_rank)
				{
					my_rank = omp_get_thread_num();
					LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
					lisqp_zero(&tmpm);
					for(i=is;i<ie;i++)
					{
						lisqp_mul_dd_d(&xy,&ym[i],x[i]);
						lisqp_add(&tmpm,&tmpm,&xy);
					}
					((LIS_MSCALAR *)lis_vec_tmp)[my_rank*LIS_VEC_TMP_PADD] = tmpm;
				}
				lisqp_zero(&dotm);
				for(i=0;i<nprocs;i++)
				{
					lisqp_add(&dotm,&dotm,&((LIS_MSCALAR *)lis_vec_tmp)[i*LIS_VEC_TMP_PADD]);
				}
			#else
				lisqp_zero(&dotm);
				for(i=0;i<n;i++)
				{
					lisqp_mul_dd_d(&xy,&ym[i],x[i]);
					lisqp_add(&dotm,&dotm,&xy);
				}
			#endif
			#ifdef USE_MPI
				MPI_Allreduce(&dotm,&tmpm,1,LIS_MPI_MSCALAR,LIS_MPI_MSUM,comm);
				*val = tmpm;
			#else
				*val = dotm;
			#endif
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}




#undef __FUNC__
#define __FUNC__ "lis_vector_nrm2ex"
int lis_vector_nrm2ex(LIS_VECTOR vx, LIS_MSCALAR *val)
{
	int i,n,is,ie,nprocs,my_rank;
	LIS_SCALAR dot,tmp;
	LIS_SCALAR *x;
	LIS_MSCALAR dotm,tmpm,xx;
	LIS_MSCALAR *xm;
	#ifdef USE_MPI
		MPI_Comm   comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	n      = vx->n;

	if( vx->precision!=LIS_PRECISION_DOUBLE )
	{
		xm = (LIS_MSCALAR *)vx->value;
		#ifdef USE_MPI
			comm   = vx->comm;
		#endif
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			#pragma omp parallel private(i,tmpm,xx,my_rank)
			{
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
				lisqp_zero(&tmpm);
				for(i=is;i<ie;i++)
				{
					lisqp_sqr(&xx,&xm[i]);
					lisqp_add(&tmpm,&tmpm,&xx);
				}
				((LIS_MSCALAR *)lis_vec_tmp)[my_rank*LIS_VEC_TMP_PADD] = tmpm;
			}
			lisqp_zero(&dotm);
			for(i=0;i<nprocs;i++)
			{
				lisqp_add(&dotm,&dotm,&((LIS_MSCALAR *)lis_vec_tmp)[i*LIS_VEC_TMP_PADD]);
			}
		#else
				lisqp_zero(&dotm);
				for(i=0;i<n;i++)
				{
					lisqp_sqr(&xx,&xm[i]);
					lisqp_add(&dotm,&dotm,&xx);
				}
		#endif
		#ifdef USE_MPI
			MPI_Allreduce(&dotm,&tmpm,1,LIS_MPI_MSCALAR,LIS_MPI_MSUM,comm);
			lisqp_sqrt(val,&tmpm);
		#else
			lisqp_sqrt(val,&dotm);
		#endif
	}
	else
	{
		x   = vx->value;
		#ifdef USE_MPI
			comm   = vx->comm;
		#endif
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			#pragma omp parallel private(i,tmp,my_rank)
			{
				my_rank = omp_get_thread_num();
				LIS_GET_ISIE(my_rank,nprocs,n,is,ie);
				tmp     = (LIS_SCALAR)0.0;
				#pragma omp for
				for(i=0;i<n;i++)
				{
					tmp += x[i]*x[i];
				}
				lis_vec_tmp[my_rank*LIS_VEC_TMP_PADD] = tmp;
			}
			dot = 0.0;
			for(i=0;i<nprocs;i++)
			{
				dot += lis_vec_tmp[i*LIS_VEC_TMP_PADD];
			}
		#else
			dot = 0.0;
			for(i=0; i<n; i++)
			{
				dot += x[i]*x[i];
			}
		#endif
		#ifdef USE_MPI
			MPI_Allreduce(&dot,&tmp,1,MPI_DOUBLE,MPI_SUM,comm);
			(*val).hi = sqrt(tmp);
		#else
			(*val).hi = sqrt(dot);
		#endif
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


#undef __FUNC__
#define __FUNC__ "lis_vector_sumex"
int lis_vector_sumex(LIS_VECTOR vx, LIS_MSCALAR *val)
{
	int i,n,nprocs,my_rank;
	LIS_SCALAR sum,tmp;
	LIS_SCALAR *x;
	LIS_MSCALAR dotm,tmpm,summ;
	LIS_MSCALAR *xm;
	#ifdef USE_MPI
		MPI_Comm   comm;
	#endif

	LIS_DEBUG_FUNC_IN;

	n      = vx->n;

	if( vx->precision!=LIS_PRECISION_DOUBLE )
	{
		xm = (LIS_MSCALAR *)vx->value;
		#ifdef USE_MPI
			comm   = vx->comm;
		#endif
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			#pragma omp parallel private(i,tmpm,my_rank)
			{
				my_rank = omp_get_thread_num();
				lisqp_zero(&tmpm);
				#pragma omp for
				for(i=0;i<n;i++)
				{
					lisqp_add(&tmpm,&tmpm,&xm[i]);
				}
				((LIS_MSCALAR *)lis_vec_tmp)[my_rank*LIS_VEC_TMP_PADD] = tmpm;
			}
			lisqp_zero(&summ);
			for(i=0;i<nprocs;i++)
			{
				lisqp_add(&summ,&summ,&((LIS_MSCALAR *)lis_vec_tmp)[i*LIS_VEC_TMP_PADD]);
			}
		#else
			lisqp_zero(&summ);
			for(i=0; i<n; i++)
			{
				lisqp_add(&summ,&summ,&xm[i]);
			}
		#endif
		#ifdef USE_MPI
			MPI_Allreduce(&summ,&tmpm,1,LIS_MPI_MSCALAR,LIS_MPI_MSUM,comm);
			*val = tmpm;
		#else
			*val = summ;
		#endif
	}
	else
	{
		x   = vx->value;
		#ifdef USE_MPI
			comm   = vx->comm;
		#endif
		#ifdef _OPENMP
			nprocs = omp_get_max_threads();
			#pragma omp parallel private(i,tmp,my_rank)
			{
				my_rank = omp_get_thread_num();
				tmp     = (LIS_SCALAR)0.0;
				#pragma omp for
				for(i=0;i<n;i++)
				{
					tmp += x[i];
				}
				lis_vec_tmp[my_rank*LIS_VEC_TMP_PADD] = tmp;
			}
			sum = 0.0;
			for(i=0;i<nprocs;i++)
			{
				sum += lis_vec_tmp[i*LIS_VEC_TMP_PADD];
			}
		#else
			sum = 0.0;
			for(i=0; i<n; i++)
			{
				sum += x[i];
			}
		#endif
		#ifdef USE_MPI
			MPI_Allreduce(&sum,&tmp,1,MPI_DOUBLE,MPI_SUM,comm);
			(*val).hi = tmp;
		#else
			(*val).hi = sum;
		#endif
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif
