/* Copyright (C) 2004-2006 The SSI Project. All rights reserved.
   http://ssi.is.s.u-tokyo.ac.jp/

   This software is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY.  See the LICENSE file for the precise
   terms and conditions for the use of this software.
*/
#ifdef HAVE_CONFIG_H
	#include "config.h"
#else
#ifdef HAVE_CONFIG_WIN32_H
	#include "config_win32.h"
#endif
#endif

#include <math.h>
#ifdef _OPENMP
	#include <omp.h>
#endif
#ifdef USE_MPI
	#include <mpi.h>
#endif
#include "lislib.h"

/************************************************
 * lis_vector_axpy			y   <- y + alpha * x
 * lis_vector_xpay			y   <- x + alpha * y
 * lis_vector_axpyz			z   <- y + alpha * x
 * lis_vector_copy			y   <- x
 * lis_vector_scale			y   <- alpha * x
 * lis_vector_pmul			z_i <- x_i * y_i
 * lis_vector_pdiv			z_i <- x_i / y_i
 * lis_vector_set_all		x_i <- alpha
 * lis_vector_abs			x_i <- |x_i|
 * lis_vector_reciprocal	x_i <- 1 / x_i
 * lis_vector_shift			x_i <- alpha + x_i
 ************************************************/

/**********************/
/* y <- y + alpha * x */
/**********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_axpy"
int lis_vector_axpy(LIS_SCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy)
{
	int i,n;
	LIS_SCALAR *x,*y;

	LIS_DEBUG_FUNC_IN;

	n    = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x    = vx->value;
	y    = vy->value;
	#pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] += alpha * x[i];
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


/**********************/
/* y <- x + alpha * y */
/**********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_xpay"
int lis_vector_xpay(LIS_VECTOR vx, LIS_SCALAR alpha, LIS_VECTOR vy)
{
	int i,n;
	LIS_SCALAR *x,*y;

	LIS_DEBUG_FUNC_IN;

	n    = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x    = vx->value;
	y    = vy->value;
	#pragma omp parallel for private(i)
    #pragma cdir nodep
	for(i=0; i<n; i++)
	{
		y[i] = x[i] + alpha * y[i];
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}



/**********************/
/* z <- y + alpha * x */
/**********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_axpyz"
int lis_vector_axpyz(LIS_SCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy, LIS_VECTOR vz)
{
	int i,n;
	LIS_SCALAR *x,*y,*z;

	LIS_DEBUG_FUNC_IN;

	n    = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n || n!=vz->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y and z is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x    = vx->value;
	y    = vy->value;
	z    = vz->value;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		z[i] = alpha * x[i] + y[i];
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


/********************/
/* y <- x           */
/********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_copy"
int lis_vector_copy(LIS_VECTOR vx, LIS_VECTOR vy)
{
	int i,n;
	LIS_SCALAR *x,*y;

	LIS_DEBUG_FUNC_IN;

	n = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x = vx->value;
	y = vy->value;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		y[i] = x[i];
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


/********************/
/* y <- alpha * x   */
/********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_scale"
int lis_vector_scale(LIS_SCALAR alpha, LIS_VECTOR vx)
{
	int i,n;
	LIS_SCALAR *x;

	LIS_DEBUG_FUNC_IN;

	n = vx->n;

	x = vx->value;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		x[i] = alpha * x[i];
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


/********************/
/* x_i <- alpha     */
/********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_set_all"
int lis_vector_set_all(LIS_SCALAR alpha, LIS_VECTOR vx)
{
	int i,n;
	LIS_SCALAR *x;

	LIS_DEBUG_FUNC_IN;

	n  = vx->n;

	x  = vx->value;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		x[i] = alpha;
	}

	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}


/********************/
/* z_i <- x_i * y_i */
/********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_pmul"
int lis_vector_pmul(LIS_VECTOR vx,LIS_VECTOR vy,LIS_VECTOR vz)
{
	int i,n;
	LIS_SCALAR *x,*y,*z;

	LIS_DEBUG_FUNC_IN;

	n    = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n || n!=vz->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y and z is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x    = vx->value;
	y    = vy->value;
	z    = vz->value;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		z[i] = x[i] * y[i];
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

/********************/
/* z_i <- x_i / y_i */
/********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_pdiv"
int lis_vector_pdiv(LIS_VECTOR vx,LIS_VECTOR vy,LIS_VECTOR vz)
{
	int i,n;
	LIS_SCALAR *x,*y,*z;

	LIS_DEBUG_FUNC_IN;

	n    = vx->n;	
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n || n!=vz->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y and z is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x    = vx->value;
	y    = vy->value;
	z    = vz->value;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		z[i] = x[i] / y[i];
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

/********************/
/* x_i <- |x_i|     */
/********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_abs"
int lis_vector_abs(LIS_VECTOR vx)
{
	int i,n;
	LIS_SCALAR *x;

	LIS_DEBUG_FUNC_IN;

	x = vx->value;
	n = vx->n;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		x[i] = fabs(x[i]);
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

/********************/
/* x_i <- 1 / x_i   */
/********************/
#undef __FUNC__
#define __FUNC__ "lis_vector_reciprocal"
int lis_vector_reciprocal(LIS_VECTOR vx)
{
	int i,n;
	LIS_SCALAR *x;

	LIS_DEBUG_FUNC_IN;

	x = vx->value;
	n = vx->n;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		x[i] = 1.0 / x[i];
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

/************************/
/* x_i <- alpha + x_i   */
/************************/
#undef __FUNC__
#define __FUNC__ "lis_vector_shift"
int lis_vector_shift(LIS_SCALAR alpha, LIS_VECTOR vx)
{
	int i,n;
	LIS_SCALAR *x;

	LIS_DEBUG_FUNC_IN;

	x = vx->value;
	n = vx->n;
    #pragma cdir nodep
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		x[i] = alpha + x[i];
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#ifdef USE_QUAD_PRECISION
#undef __FUNC__
#define __FUNC__ "lis_vector_axpyex"
int lis_vector_axpyex(LIS_MSCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy)
{
	int i,n,is,ie,nprocs,my_rank;
	LIS_SCALAR *x;
	LIS_MSCALAR *xm,*ym,ax;

	LIS_DEBUG_FUNC_IN;

	n    = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
		if( vy->precision==LIS_PRECISION_DOUBLE )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"third argument is not multi precision vector\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	if( vx->precision!=LIS_PRECISION_DOUBLE )
	{
		xm   = (LIS_MSCALAR *)vx->value;
		ym   = (LIS_MSCALAR *)vy->value;
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			lisqp_mul(&ax,&alpha,&xm[i]);
			lisqp_add(&ym[i],&ym[i],&ax);
		}
	}
	else
	{
		x    = vx->value;
		ym   = (LIS_MSCALAR *)vy->value;
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			lisqp_mul_dd_d(&ax,&alpha,x[i]);
			lisqp_add(&ym[i],&ym[i],&ax);
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_vector_xpayex"
int lis_vector_xpayex(LIS_VECTOR vx, LIS_MSCALAR alpha, LIS_VECTOR vy)
{
	int i,n,is,ie,nprocs,my_rank;
	LIS_SCALAR *x;
	LIS_MSCALAR *xm,*ym,ay;

	LIS_DEBUG_FUNC_IN;

	n    = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
		if( vy->precision==LIS_PRECISION_DOUBLE )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"vector y is not multi precision vector\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	if( vx->precision!=LIS_PRECISION_DOUBLE )
	{
		xm   = (LIS_MSCALAR *)vx->value;
		ym   = (LIS_MSCALAR *)vy->value;
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			lisqp_mul(&ay,&alpha,&ym[i]);
			lisqp_add(&ym[i],&xm[i],&ay);
		}
	}
	else
	{
		x    = vx->value;
		ym   = (LIS_MSCALAR *)vy->value;
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			lisqp_mul(&ay,&alpha,&ym[i]);
			lisqp_add_dd_d(&ym[i],&ay,x[i]);
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_vector_axpyzex"
int lis_vector_axpyzex(LIS_MSCALAR alpha, LIS_VECTOR vx, LIS_VECTOR vy, LIS_VECTOR vz)
{
	int i,n,is,ie,nprocs,my_rank;
	LIS_SCALAR *x,*y;
	LIS_MSCALAR *xm,*ym,*zm,ax;

	LIS_DEBUG_FUNC_IN;

	n    = vx->n;

	#ifndef NO_ERROR_CHECK
		if( n!=vy->n || n!=vz->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y and z is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
		if( vz->precision==LIS_PRECISION_DOUBLE )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"vector z is not multi precision vector\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	if( vx->precision>=LIS_PRECISION_QUAD && vy->precision>=LIS_PRECISION_QUAD )
	{
		xm   = (LIS_MSCALAR *)vx->value;
		ym   = (LIS_MSCALAR *)vy->value;
		zm   = (LIS_MSCALAR *)vz->value;
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			lisqp_mul(&ax,&alpha,&xm[i]);
			lisqp_add(&zm[i],&ym[i],&ax);
		}
	}
	else if( vx->precision==LIS_PRECISION_DEFAULT && vy->precision==LIS_PRECISION_DEFAULT )
	{
		x    = vx->value;
		y    = vy->value;
		zm   = (LIS_MSCALAR *)vz->value;
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			lisqp_mul_dd_d(&ax,&alpha,x[i]);
			lisqp_add_dd_d(&zm[i],&ax,y[i]);
		}
	}
	else
	{
		if( vx->precision!=LIS_PRECISION_DOUBLE )
		{
			xm   = (LIS_MSCALAR *)vx->value;
			y    = vy->value;
			zm   = (LIS_MSCALAR *)vz->value;
			#pragma omp parallel for private(i)
			for(i=0; i<n; i++)
			{
				lisqp_mul(&ax,&alpha,&xm[i]);
				lisqp_add_dd_d(&zm[i],&ax,y[i]);
			}
		}
		else
		{
			x    = vx->value;
			ym   = (LIS_MSCALAR *)vy->value;
			zm   = (LIS_MSCALAR *)vz->value;
			#pragma omp parallel for private(i)
			for(i=0; i<n; i++)
			{
				lisqp_mul_dd_d(&ax,&alpha,x[i]);
				lisqp_add(&zm[i],&ym[i],&ax);
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_vector_copyex"
int lis_vector_copyex(LIS_VECTOR vx, LIS_VECTOR vy)
{
	int i,n;
	LIS_SCALAR *x,*y;
	LIS_MSCALAR *xm,*ym;

	LIS_DEBUG_FUNC_IN;

	n = vx->n;
	#ifndef NO_ERROR_CHECK
		if( n!=vy->n )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"length of vector x and y is not equal\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	if( vx->precision>=LIS_PRECISION_QUAD && vy->precision>=LIS_PRECISION_QUAD )
	{
		xm   = (LIS_MSCALAR *)vx->value;
		ym   = (LIS_MSCALAR *)vy->value;
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			ym[i] = xm[i];
		}
	}
	else if( vx->precision==LIS_PRECISION_DEFAULT && vy->precision==LIS_PRECISION_DEFAULT )
	{
		x   = vx->value;
		y   = vy->value;
		#pragma omp parallel for private(i)
		for(i=0; i<n; i++)
		{
			y[i] = x[i];
		}
	}
	else
	{
		if( vx->precision!=LIS_PRECISION_DOUBLE )
		{
			xm   = (LIS_MSCALAR *)vx->value;
			y    = vy->value;
			#pragma omp parallel for private(i)
			for(i=0; i<n; i++)
			{
				y[i] = xm[i].hi;
			}
		}
		else
		{
			x    = vx->value;
			ym   = (LIS_MSCALAR *)vy->value;
			#pragma omp parallel for private(i)
			for(i=0; i<n; i++)
			{
				ym[i].hi = x[i];
				ym[i].lo = 0.0;
			}
		}
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_vector_scaleex"
int lis_vector_scaleex(LIS_MSCALAR alpha, LIS_VECTOR vx)
{
	int i,n,is,ie,nprocs,my_rank;
	LIS_MSCALAR *x;

	LIS_DEBUG_FUNC_IN;

	n = vx->n;
	#ifndef NO_ERROR_CHECK
		if( vx->precision==LIS_PRECISION_DOUBLE )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"vector x is not multi precision vector\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x = (LIS_MSCALAR *)vx->value;
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		lisqp_mul(&x[i],&alpha,&x[i]);
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}

#undef __FUNC__
#define __FUNC__ "lis_vector_set_allex"
int lis_vector_set_allex(LIS_MSCALAR alpha, LIS_VECTOR vx)
{
	int i,n;
	LIS_MSCALAR *x;

	LIS_DEBUG_FUNC_IN;

	n  = vx->n;
	#ifndef NO_ERROR_CHECK
		if( vx->precision==LIS_PRECISION_DOUBLE )
		{
			LIS_SETERR(LIS_ERR_ILL_ARG,"vector x is not multi precision vector\n");
			return LIS_ERR_ILL_ARG;
		}
	#endif

	x  = (LIS_MSCALAR *)vx->value;
	#pragma omp parallel for private(i)
	for(i=0; i<n; i++)
	{
		x[i] = alpha;
	}
	LIS_DEBUG_FUNC_OUT;
	return LIS_SUCCESS;
}
#endif
